# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2022 by the BurnMan team, released under the GNU
# GPL v2 or later.


from __future__ import absolute_import

import numpy as np
from sympy import Matrix, nsimplify
import scipy.optimize as opt

from burnman.classes.solutionmodel import IdealSolution

from ..constants import gas_constant
from .material import material_property, cached_property
from .mineral import Mineral
from .elasticsolutionmodel import ElasticMechanicalSolution
from .averaging_schemes import reuss_average_function

from ..utils.math import bracket
from ..utils.reductions import independent_row_indices
from ..utils.chemistry import sum_formulae, sort_element_list_to_IUPAC_order


class ElasticSolution(Mineral):
    """
    This is the base class for all Elastic solutions.
    Site occupancies, endmember activities and the constant
    and volume and temperature dependencies of the excess
    properties can be queried after using set_composition().
    States of the solution can only be queried after setting
    the pressure, temperature and composition using set_state()
    and set_composition.

    This class is available as :class:`burnman.ElasticSolution`.
    It uses an instance of :class:`burnman.ElasticSolutionModel` to
    calculate interaction terms between endmembers.

    All the solution parameters are expected to be in SI units.  This
    means that the interaction parameters should be in J/mol, with the T
    and V derivatives in J/K/mol and Pa/mol.

    The parameters are relevant to all Elastic solution models. Please
    see the documentation for individual models for details about
    other parameters.

    :param name: Name of the solution.
    :type name: string
    :param solution_model: The ElasticSolutionModel object defining the
        properties of the solution.
    :type solution_model: :class:`burnman.ElasticSolutionModel`
    :param molar_fractions: The molar fractions of each endmember in the solution.
        Can be reset using the set_composition() method.
    :type molar_fractions: numpy.array
    """

    def __init__(self, name=None, solution_model=None, molar_fractions=None):
        """
        Set up matrices to speed up calculations for when P, T, X is defined.
        """
        Mineral.__init__(self)

        # Solution needs a method attribute to call Mineral.set_state().
        # Note that set_method() below will not change self.method
        self.method = "ElasticSolutionMethod"

        if name is not None:
            self.name = name
        if solution_model is not None:
            self.solution_model = solution_model

        if isinstance(solution_model, ElasticMechanicalSolution):
            self.solution_type = "mechanical"
        else:
            self.solution_type = "chemical"

        # Starting guess and delta for pressure iteration
        self.min_V0 = min(
            [mbr[0].params["V_0"] for mbr in self.solution_model.endmembers]
        )
        self.dV = 0.01 * self.min_V0

        # Equation of state
        for i in range(self.n_endmembers):
            self.solution_model.endmembers[i][0].set_method(
                self.solution_model.endmembers[i][0].params["equation_of_state"]
            )

        # Molar fractions
        if molar_fractions is not None:
            self.set_composition(molar_fractions)

    @cached_property
    def endmembers(self):
        return self.solution_model.endmembers

    def set_composition(self, molar_fractions):
        """
        Set the composition for this solution.
        Resets cached properties.

        :param molar_fractions: Molar abundance for each endmember,
            needs to sum to one.
        :type molar_fractions: list of float
        """
        assert len(self.solution_model.endmembers) == len(molar_fractions)

        if self.solution_type != "mechanical":
            assert sum(molar_fractions) > 0.9999
            assert sum(molar_fractions) < 1.0001

        self.reset()
        self.molar_fractions = np.array(molar_fractions)

        if self.temperature is not None:
            _ = self.molar_volume

    def set_method(self, method):
        for i in range(self.n_endmembers):
            self.solution_model.endmembers[i][0].set_method(method)
        # note: do not set self.method here!
        self.reset()

    def set_state(self, pressure, temperature):
        Mineral.set_state(self, pressure, temperature)

        try:
            _ = self.molar_volume
        except AttributeError:
            pass

    @material_property
    def formula(self):
        """
        Returns molar chemical formula of the solution.
        """
        return sum_formulae(self.endmember_formulae, self.molar_fractions)

    @material_property
    def activities(self):
        """
        Returns a list of endmember activities [unitless].
        """
        volumes = [
            self.solution_model.endmembers[i][0].method.volume(
                self.pressure,
                self.temperature,
                self.solution_model.endmembers[i][0].params,
            )
            for i in range(self.n_endmembers)
        ]

        gibbs_pure = [
            self.solution_model.endmembers[i][0].method.gibbs_free_energy(
                self.pressure,
                self.temperature,
                volumes[i],
                self.solution_model.endmembers[i][0].params,
            )
            for i in range(self.n_endmembers)
        ]

        acts = np.exp(
            (self.partial_gibbs - np.array(gibbs_pure))
            / (gas_constant * self.temperature)
        )
        return acts

    @material_property
    def activity_coefficients(self):
        """
        Returns a list of endmember activity coefficients
        (gamma = activity / ideal activity) [unitless].
        """
        return np.exp(
            np.log(self.activities)
            - IdealSolution._log_ideal_activities(
                self.solution_model, self.molar_fractions
            )
        )

    @material_property
    def molar_internal_energy(self):
        """
        Returns molar internal energy of the mineral [J/mol].
        Aliased with self.energy
        """
        return self.molar_helmholtz + self.temperature * self.molar_entropy

    @material_property
    def _excess_partial_helmholtz(self):
        """
        Returns excess partial molar helmholtz energy
        at constant volume [J/mol].
        Property specific to solutions.
        """
        return self.solution_model.excess_partial_helmholtz_energies(
            self.molar_volume, self.temperature, self.molar_fractions
        )

    @material_property
    def _excess_partial_pressures(self):
        """
        Returns excess partial pressures at constant volume [Pa].
        Property specific to solutions.
        """
        return self.solution_model.excess_partial_pressures(
            self.molar_volume, self.temperature, self.molar_fractions
        )

    @material_property
    def _excess_partial_entropies(self):
        """
        Returns excess partial entropies at constant volume [J/K].
        Property specific to solutions.
        """
        return self.solution_model.excess_partial_entropies(
            self.molar_volume, self.temperature, self.molar_fractions
        )

    @material_property
    def _partial_helmholtz(self):
        """
        Returns endmember partial molar Helmholtz energy at constant volume [J/mol].
        Property specific to solutions.
        """
        return (
            np.array(
                [
                    self.solution_model.endmembers[i][0].helmholtz
                    for i in range(self.n_endmembers)
                ]
            )
            + self._excess_partial_helmholtz
        )

    @material_property
    def _partial_pressures(self):
        """
        Returns endmember partial pressures at constant volume [Pa].
        Property specific to solutions.
        """
        return (
            np.array(
                [
                    self.solution_model.endmembers[i][0].pressure
                    for i in range(self.n_endmembers)
                ]
            )
            + self._excess_partial_pressures
        )

    @material_property
    def _partial_entropies(self):
        """
        Returns endmember partial entropies at constant volume [J/K].
        Property specific to solutions.
        """
        return (
            np.array(
                [
                    self.solution_model.endmembers[i][0].molar_entropy
                    for i in range(self.n_endmembers)
                ]
            )
            + self._excess_partial_entropies
        )

    @material_property
    def partial_gibbs(self):
        """
        Returns endmember partial molar Gibbs energy
        at constant pressure [J/mol].
        Property specific to solutions.
        """
        return self._partial_helmholtz + self.pressure * self.molar_volume

    @material_property
    def _dPdX(self):
        """
        Returns the change in pressure with amount of each endmember
        at constant volume.
        """
        sumX = np.sum(self.molar_fractions)
        sumXP = np.einsum("i,i->", self.molar_fractions, self._partial_pressures)
        return (self._partial_pressures * sumX - sumXP) / (sumX * sumX)

    @material_property
    def _dVdX(self):
        """
        Returns the change in pressure with amount of each endmember
        at constant pressure.
        """
        return self.molar_volume / self.isothermal_bulk_modulus_reuss * self._dPdX

    @material_property
    def _dSdX_mod(self):
        """
        Returns the additional change in entropy with
        amount of each endmember due to converting from constant volume
        to constant pressure
        """
        return self.alpha * self.molar_volume * self._dPdX

    @material_property
    def partial_volumes(self):
        """
        Returns endmember partial molar volumes [m^3/mol].
        Property specific to solutions.
        """
        A = np.eye(self.n_endmembers) - self.molar_fractions
        Vs = self.molar_volume + np.einsum("ij, j->i", A, self._dVdX)
        return Vs

    @material_property
    def partial_entropies(self):
        """
        Returns endmember partial molar entropies [J/K/mol].
        Property specific to solutions.
        """
        A = np.eye(self.n_endmembers) - self.molar_fractions
        Ss = self._partial_entropies + np.einsum("ij, j->i", A, self._dSdX_mod)
        return Ss

    @material_property
    def _excess_helmholtz(self):
        """
        Returns molar excess Helmholtz energy at constant volume [J/mol].
        Property specific to solutions.
        """
        return self.solution_model.excess_helmholtz_energy(
            self.molar_volume, self.temperature, self.molar_fractions
        )

    @material_property
    def _helmholtz_hessian(self):
        """
        Returns an array containing the second compositional derivative
        of the Helmholtz energy at constant volume [J/mol].
        Property specific to solutions.
        """
        return self.solution_model.helmholtz_hessian(
            self.molar_volume, self.temperature, self.molar_fractions
        )

    @material_property
    def _entropy_hessian(self):
        """
        Returns an array containing the second compositional derivative
        of the entropy at constant volume [J/K].
        Property specific to solutions.
        """
        return self.solution_model.entropy_hessian(
            self.molar_volume, self.temperature, self.molar_fractions
        )

    @material_property
    def _pressure_hessian(self):
        """
        Returns an array containing the second compositional derivative
        of the pressure at constant volume [Pa].
        Property specific to solutions.
        """
        return self.solution_model.pressure_hessian(
            self.molar_volume, self.temperature, self.molar_fractions
        )

    @material_property
    def gibbs_hessian(self):
        """
        Returns an array containing the second compositional derivative
        of the Gibbs energy at constant pressure [J/mol].
        Property specific to solutions.
        """
        raise NotImplementedError

    @material_property
    def molar_helmholtz(self):
        """
        Returns molar Helmholtz energy of the solution [J/mol].
        Aliased with self.helmholtz.
        """
        return (
            sum(
                [
                    self.solution_model.endmembers[i][0].molar_helmholtz
                    * self.molar_fractions[i]
                    for i in range(self.n_endmembers)
                ]
            )
            + self._excess_helmholtz
        )

    @material_property
    def molar_gibbs(self):
        """
        Returns molar Gibbs free energy of the solution [J/mol].
        Aliased with self.gibbs.
        """
        return self.molar_helmholtz + self.pressure * self.molar_volume

    @material_property
    def molar_mass(self):
        """
        Returns molar mass of the solution [kg/mol].
        """
        return sum(
            [
                self.solution_model.endmembers[i][0].molar_mass
                * self.molar_fractions[i]
                for i in range(self.n_endmembers)
            ]
        )

    @material_property
    def excess_pressure(self):
        """
        Returns excess pressure of the solution [Pa].
        Specific property for solutions.
        """
        return self.solution_model.excess_pressure(
            self.molar_volume, self.temperature, self.molar_fractions
        )

    @material_property
    def molar_volume(self):
        """
        Returns molar volume of the solution [m^3/mol].
        Aliased with self.V.
        """

        def _delta_pressure(volume):
            self._ptmp = [
                self.solution_model.endmembers[i][0].method.pressure(
                    self.temperature,
                    volume,
                    self.solution_model.endmembers[i][0].params,
                )
                for i in range(self.n_endmembers)
            ]

            pressure_try = sum(
                [
                    self._ptmp[i] * self.molar_fractions[i]
                    for i in range(self.n_endmembers)
                ]
            ) + self.solution_model.excess_pressure(
                volume, self.temperature, self.molar_fractions
            )

            return pressure_try - self.pressure

        def _K_T(volume):
            # Note, this only works when the excess pressure is not a function
            # of V or T.

            return sum(
                [
                    self.solution_model.endmembers[i][
                        0
                    ].method.isothermal_bulk_modulus_reuss(
                        0.0,
                        self.temperature,
                        volume,
                        self.solution_model.endmembers[i][0].params,
                    )
                    * self.molar_fractions[i]
                    for i in range(self.n_endmembers)
                ]
            )

        try:
            # The first attempt to find a bracket for
            # root finding uses V_0 as a starting point
            sol = bracket(_delta_pressure, self.min_V0, self.dV)
        except Exception:
            # At high temperature, the naive bracketing above may
            # try a volume guess that exceeds the point at which the
            # bulk modulus goes negative at that temperature.
            # In this case, we try a more nuanced approach by
            # first finding the volume at which the bulk modulus goes
            # negative, and then either (a) raising an exception if the
            # desired pressure is less than the pressure at that volume,
            # or (b) using that pressure to create a better bracket for
            # brentq.

            sol_K_T = bracket(_K_T, self.min_V0, self.dV)
            V_crit = opt.brentq(_K_T, sol_K_T[0], sol_K_T[1])
            P_min = self.pressure + _delta_pressure(V_crit)
            if P_min > self.pressure:
                raise Exception(
                    "The desired pressure is not achievable "
                    "at this temperature. The minimum pressure "
                    f"achievable is {P_min:.2e} Pa."
                )
            else:
                try:
                    sol = bracket(_delta_pressure, V_crit - self.dV, self.dV)
                except Exception:
                    raise Exception(
                        "Cannot find a volume, perhaps you are "
                        "outside of the range of validity for "
                        "the equation of state?"
                    )

        V = opt.brentq(_delta_pressure, sol[0], sol[1])

        _delta_pressure(V)
        for i in range(self.n_endmembers):
            self.solution_model.endmembers[i][0].set_state(
                self._ptmp[i], self.temperature
            )

        return V

    @material_property
    def density(self):
        """
        Returns density of the solution [kg/m^3].
        Aliased with self.rho.
        """
        return self.molar_mass / self.molar_volume

    @material_property
    def excess_entropy(self):
        """
        Returns excess molar entropy [J/K/mol].
        Property specific to solutions.
        """
        return self.solution_model.excess_entropy(
            self.molar_volume, self.temperature, self.molar_fractions
        )

    @material_property
    def molar_entropy(self):
        """
        Returns molar entropy of the solution [J/K/mol].
        Aliased with self.S.
        """
        return (
            sum(
                [
                    self.solution_model.endmembers[i][0].S * self.molar_fractions[i]
                    for i in range(self.n_endmembers)
                ]
            )
            + self.excess_entropy
        )

    @material_property
    def excess_enthalpy(self):
        """
        Returns excess molar enthalpy [J/mol].
        Property specific to solutions.
        """
        return self.solution_model.excess_enthalpy(
            self.molar_volume, self.temperature, self.molar_fractions
        )

    @material_property
    def molar_enthalpy(self):
        """
        Returns molar enthalpy of the solution [J/mol].
        Aliased with self.H.
        """
        return (
            sum(
                [
                    self.solution_model.endmembers[i][0].H * self.molar_fractions[i]
                    for i in range(self.n_endmembers)
                ]
            )
            + self.excess_enthalpy
        )

    @material_property
    def isothermal_bulk_modulus_reuss(self):
        """
        Returns isothermal bulk modulus of the solution [Pa].
        Aliased with self.K_T.
        """
        return sum(
            [
                self.solution_model.endmembers[i][0].isothermal_bulk_modulus_reuss
                * self.molar_fractions[i]
                for i in range(self.n_endmembers)
            ]
        )

    @material_property
    def isentropic_bulk_modulus_reuss(self):
        """
        Returns adiabatic bulk modulus of the solution [Pa].
        Aliased with self.K_S.
        """
        if self.temperature < 1e-10:
            return self.isothermal_bulk_modulus_reuss
        else:
            return (
                self.isothermal_bulk_modulus_reuss
                * self.molar_heat_capacity_p
                / self.molar_heat_capacity_v
            )

    @material_property
    def isothermal_compressibility_reuss(self):
        """
        Returns isothermal compressibility of the solution.
        (or inverse isothermal bulk modulus) [1/Pa].
        Aliased with self.K_T.
        """
        return 1.0 / self.isothermal_bulk_modulus_reuss

    @material_property
    def isentropic_compressibility_reuss(self):
        """
        Returns adiabatic compressibility of the solution.
        (or inverse adiabatic bulk modulus) [1/Pa].
        Aliased with self.K_S.
        """
        return 1.0 / self.isentropic_bulk_modulus_reuss

    @material_property
    def shear_modulus(self):
        """
        Returns shear modulus of the solution [Pa].
        Aliased with self.G.
        """
        G_list = np.fromiter(
            (e[0].G for e in self.solution_model.endmembers),
            dtype=float,
            count=self.n_endmembers,
        )
        return reuss_average_function(self.molar_fractions, G_list)

    @material_property
    def p_wave_velocity(self):
        """
        Returns P wave speed of the solution [m/s].
        Aliased with self.v_p.
        """
        return np.sqrt(
            (self.isentropic_bulk_modulus_reuss + 4.0 / 3.0 * self.shear_modulus)
            / self.density
        )

    @material_property
    def bulk_sound_velocity(self):
        """
        Returns bulk sound speed of the solution [m/s].
        Aliased with self.v_phi.
        """
        return np.sqrt(self.isentropic_bulk_modulus_reuss / self.density)

    @material_property
    def shear_wave_velocity(self):
        """
        Returns shear wave speed of the solution [m/s].
        Aliased with self.v_s.
        """
        return np.sqrt(self.shear_modulus / self.density)

    @material_property
    def grueneisen_parameter(self):
        """
        Returns grueneisen parameter of the solution [unitless].
        Aliased with self.gr.
        """
        if self.temperature < 1e-10:
            return float("nan")
        else:
            return (
                self.thermal_expansivity
                * self.isothermal_bulk_modulus_reuss
                * self.molar_volume
                / self.molar_heat_capacity_v
            )

    @material_property
    def thermal_expansivity(self):
        """
        Returns thermal expansion coefficient (alpha)
        of the solution [1/K].
        Aliased with self.alpha.
        """
        alphaKT = sum(
            [
                self.solution_model.endmembers[i][0].isothermal_bulk_modulus_reuss
                * self.solution_model.endmembers[i][0].alpha
                * self.molar_fractions[i]
                for i in range(self.n_endmembers)
            ]
        )
        return alphaKT / self.isothermal_bulk_modulus_reuss

    @material_property
    def molar_heat_capacity_v(self):
        """
        Returns molar heat capacity at constant volume of the
        solution [J/K/mol].
        Aliased with self.C_v.
        """
        return sum(
            [
                self.solution_model.endmembers[i][0].molar_heat_capacity_v
                * self.molar_fractions[i]
                for i in range(self.n_endmembers)
            ]
        )

    @material_property
    def molar_heat_capacity_p(self):
        """
        Returns molar heat capacity at constant pressure
        of the solution [J/K/mol].
        Aliased with self.C_p.
        """
        return (
            self.molar_heat_capacity_v
            + self.molar_volume
            * self.temperature
            * self.thermal_expansivity
            * self.thermal_expansivity
            * self.isothermal_bulk_modulus_reuss
        )

    @cached_property
    def stoichiometric_matrix(self):
        """
        A sympy Matrix where each element M[i,j] corresponds
        to the number of atoms of element[j] in endmember[i].
        """

        def f(i, j):
            e = self.elements[j]
            if e in self.endmember_formulae[i]:
                return nsimplify(self.endmember_formulae[i][e])
            else:
                return 0

        return Matrix(len(self.endmember_formulae), len(self.elements), f)

    @cached_property
    def stoichiometric_array(self):
        """
        An array where each element arr[i,j] corresponds
        to the number of atoms of element[j] in endmember[i].
        """
        return np.array(self.stoichiometric_matrix)

    @cached_property
    def reaction_basis(self):
        """
        An array where each element arr[i,j] corresponds
        to the number of moles of endmember[j] involved in reaction[i].
        """
        reaction_basis = np.array(
            [v[:] for v in self.stoichiometric_matrix.T.nullspace()]
        )

        if len(reaction_basis) == 0:
            reaction_basis = np.empty((0, len(self.endmember_names)))

        return reaction_basis

    @cached_property
    def n_reactions(self):
        """
        The number of reactions in reaction_basis.
        """
        return len(self.reaction_basis[:, 0])

    @cached_property
    def independent_element_indices(self):
        """
        A list of an independent set of element indices. If the amounts of
        these elements are known (element_amounts),
        the amounts of the other elements can be inferred by
        -compositional_null_basis[independent_element_indices].dot(element_amounts).
        """
        return sorted(independent_row_indices(self.stoichiometric_matrix.T))

    @cached_property
    def dependent_element_indices(self):
        """
        The element indices not included in the independent list.
        """
        return [
            i
            for i in range(len(self.elements))
            if i not in self.independent_element_indices
        ]

    @cached_property
    def compositional_null_basis(self):
        """
        An array N such that N.b = 0 for all bulk compositions that can
        be produced with a linear sum of the endmembers in the solution.
        """
        null_basis = np.array([v[:] for v in self.stoichiometric_matrix.nullspace()])

        M = null_basis[:, self.dependent_element_indices]
        assert (M.shape[0] == M.shape[1]) and (M == np.eye(M.shape[0])).all()

        return null_basis

    @cached_property
    def endmember_formulae(self):
        """
        A list of formulae for all the endmember in the solution.
        """
        return [mbr[0].params["formula"] for mbr in self.solution_model.endmembers]

    @cached_property
    def endmember_names(self):
        """
        A list of names for all the endmember in the solution.
        """
        return [mbr[0].name for mbr in self.solution_model.endmembers]

    @cached_property
    def n_endmembers(self):
        """
        The number of endmembers in the solution.
        """
        return len(self.solution_model.endmembers)

    @cached_property
    def elements(self):
        """
        A list of the elements which could be contained in the solution,
        returned in the IUPAC element order.
        """
        keys = []
        for f in self.endmember_formulae:
            keys.extend(f.keys())

        return sort_element_list_to_IUPAC_order(set(keys))


ElasticSolidSolution = ElasticSolution



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2024 by the BurnMan team, released under the GNU
# GPL v2 or later.

"""
Decker_1971
^^^^^^^^^^^
"""

from burnman.eos.mie_grueneisen_debye import MGDBase
from burnman.classes.calibrant import Calibrant
from burnman.eos.birch_murnaghan_4th import pressure_birch_murnaghan_fourth


class NaCl_B1(Calibrant):
    """
    The NaCl (B1 structured) pressure standard reported by
    Decker (1971; https://doi.org/10.1063/1.1660714).

    .. note:: This calibrant is not exactly the same as that proposed by Decker.
        The cold compression curve has here been approximated by a 4th order
        Birch-Murnaghan EoS, as described in
        Matsui et al. (2012; https://doi.org/10.2138/am.2012.4136).
    """

    def __init__(self):
        def _pressure_Decker_NaCl(volume, temperature, params):
            # Isothermal pressure (GPa)
            P0 = pressure_birch_murnaghan_fourth(params["V_0"] / volume, params)

            # Thermal pressure
            thermal_model = MGDBase()
            Pth0 = thermal_model._thermal_pressure(params["T_0"], volume, params)
            Pth = thermal_model._thermal_pressure(temperature, volume, params)

            return P0 + Pth - Pth0

        _params_Decker_NaCl = {
            "V_0": 2.7013e-05,
            "K_0": 23.7e9,
            "Kprime_0": 4.91,
            "Kprime_prime_0": -0.267e-9,
            "Debye_0": 279.0,
            "grueneisen_0": 1.59,
            "q_0": 0.93,
            "n": 2.0,
            "T_0": 298.15,
            "P_0": 0.0,
            "Z": 4.0,
        }

        Calibrant.__init__(self, _pressure_Decker_NaCl, "pressure", _params_Decker_NaCl)



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit
# for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.

from __future__ import absolute_import

import numpy as np
from sympy import Matrix
from scipy.linalg import block_diag

import logging
import importlib
from ..classes.polytope import MaterialPolytope, independent_row_indices
from ..classes.solution import Solution
from ..classes.composite import Composite
from .solution import transform_solution_to_new_basis

try:
    cp = importlib.import_module("cvxpy")
except ImportError as err:
    print(
        f"Warning: {err}. " "For full functionality of BurnMan, please install cvxpy."
    )


def solution_polytope_from_charge_balance(
    charges, charge_total, return_fractions=False
):
    """
    Creates a polytope object from a list of the charges for each species on
    each site and the total charge for all site-species.

    :param charges: 2D list containing the total charge for species j on site i,
        including the site multiplicity. So, for example,
        a solution with the site formula [Mg,Fe]3[Mg,Al,Si]2Si3O12 would
        have the following list: [[6., 6.], [4., 6., 8.]].
    :type charges: 2D list of floats
    :param charge_total: The total charge for all site-species per formula unit.
        The example given above would have charge_total = 12.
    :type charge_total: float
    :param return_fractions: Determines whether the created polytope object returns its
        attributes (such as endmember occupancies) as fractions or as floats.
        Default is False.
    :type return_fractions: bool

    :returns: A polytope object corresponding to the parameters provided.
    :rtype: :class:`burnman.polytope.MaterialPolytope` object
    """
    n_sites = len(charges)
    all_charges = np.concatenate(charges)
    n_site_elements = len(all_charges)
    equalities = np.empty((n_sites + 1, n_site_elements + 1))
    equalities[:-1, 0] = -1
    i = 0
    for i_site, site_charges in enumerate(charges):
        equalities[i_site, 1:] = [
            1 if (j >= i and j < i + len(site_charges)) else 0
            for j in range(n_site_elements)
        ]
        i += len(site_charges)

    equalities[-1, 0] = -charge_total
    equalities[-1, 1:] = all_charges

    pos_constraints = np.concatenate(
        (np.zeros((len(equalities[0]) - 1, 1)), np.identity(len(equalities[0]) - 1)),
        axis=1,
    )
    return MaterialPolytope(
        equalities, pos_constraints, return_fractions=return_fractions
    )


def solution_polytope_from_endmember_occupancies(
    endmember_occupancies, return_fractions=False
):
    """
    Creates a polytope object from a list of independent endmember occupancies.

    :param endmember_occupancies: 2D list containing the
        site-species occupancies j for endmember i.
        So, for example, a solution with independent endmembers
        [Mg]3[Al]2Si3O12, [Mg]3[Mg0.5Si0.5]2Si3O12, [Fe]3[Al]2Si3O12
        might have the following array:
        [[1., 0., 1., 0., 0.],
        [1., 0., 0., 0.5, 0.5],
        [0., 1., 1., 0., 0.]],
        where the order of site-species is Mg_A, Fe_A, Al_B, Mg_B, Si_B.
    :type endmember_occupancies: 2D numpy array

    :param return_fractions: Determines whether the created polytope object
        returns its attributes (such as endmember occupancies) as fractions
        or as floats.
    :type return_fractions: bool

    :returns: A polytope object corresponding to the parameters provided.
    :rtype: :class:`burnman.polytope.MaterialPolytope` object
    """
    n_sites = sum(endmember_occupancies[0])
    n_occs = endmember_occupancies.shape[1]

    nullspace = np.array(Matrix(endmember_occupancies).nullspace(), dtype=float)

    equalities = np.zeros((len(nullspace) + 1, n_occs + 1))
    equalities[0, 0] = -n_sites
    equalities[0, 1:] = 1

    if len(nullspace) > 0:
        try:
            equalities[1:, 1:] = nullspace
        except ValueError:
            equalities[1:, 1:] = nullspace[:, :, 0]

    pos_constraints = np.concatenate(
        (np.zeros((len(equalities[0]) - 1, 1)), np.identity(len(equalities[0]) - 1)),
        axis=1,
    )

    return MaterialPolytope(
        equalities,
        pos_constraints,
        return_fractions=return_fractions,
        independent_endmember_occupancies=endmember_occupancies,
    )


def composite_polytope_at_constrained_composition(
    composite, composition, return_fractions=False
):
    """
    Creates a polytope object from a Composite object and a composition.
    This polytope describes the complete set of valid composite
    endmember amounts that satisfy the compositional constraints.

    :param composite: A composite containing one or more Solution and Mineral objects.
    :type composite: :class:`burnman.Composite` object

    :param composition: A dictionary containing the amounts of each element.
    :type composition: dict

    :param return_fractions: Determines whether the created polytope object returns its
        attributes (such as endmember occupancies) as fractions or as floats.
    :type return_fractions: bool


    :returns: A polytope object corresponding to the parameters provided.
    :rtype: :class:`burnman.polytope.MaterialPolytope` object
    """
    c_array = np.empty((composite.n_elements, 1))
    c_array[:, 0] = [
        -composition[e] if e in composition else 0.0 for e in composite.elements
    ]

    equalities = np.concatenate((c_array, composite.stoichiometric_array.T), axis=1)

    eoccs = []
    for i, ph in enumerate(composite.phases):
        if isinstance(ph, Solution):
            eoccs.append(ph.solution_model.endmember_occupancies.T)
        else:
            eoccs.append(np.ones((1, 1)))

    eoccs = block_diag(*eoccs)
    inequalities = np.concatenate((np.zeros((len(eoccs), 1)), eoccs), axis=1)

    return MaterialPolytope(equalities, inequalities, return_fractions=return_fractions)


def simplify_composite_with_composition(composite, composition):
    """
    Takes a composite and a composition, and returns the simplest composite
    object that spans the solution space at the given composition.

    For example, if the composition is given as {'Mg': 2., 'Si': 1.5, 'O': 5.},
    and the composite is given as a mix of Mg,Fe olivine and pyroxene
    solutions, this function will return a composite that only contains
    the Mg-bearing endmembers.

    If the solutions have endmember proportions that are consistent with the
    bulk composition, the site occupancies of the new solution models are set to
    the values in the original models.

    :param composite: The initial Composite object.
    :type composite: :class:`burnman.Composite` object

    :param composition: A dictionary containing the amounts of each element.
    :type composition: dict

    :returns: The simplified Composite object
    :rtype: :class:`burnman.Composite` object
    """
    polytope = composite_polytope_at_constrained_composition(
        composite, composition, return_fractions=True
    )

    composite_changed = False
    new_phases = []
    mbr_amounts = polytope.endmember_occupancies
    i = 0
    for i_ph, n_mbrs in enumerate(composite.endmembers_per_phase):
        ph = composite.phases[i_ph]

        amounts = mbr_amounts[:, i : i + n_mbrs].astype(float)
        i += n_mbrs

        rank = np.linalg.matrix_rank(amounts, tol=1.0e-8)

        if rank < n_mbrs:
            if isinstance(ph, Solution) and rank > 0:
                if len(amounts) > 1:
                    c_mean = np.mean(amounts, axis=0)
                else:
                    c_mean = amounts[0]

                poly = solution_polytope_from_endmember_occupancies(
                    ph.solution_model.endmember_occupancies
                )
                dmbrs = poly.endmembers_as_independent_endmember_amounts

                x = cp.Variable(dmbrs.shape[0])
                objective = cp.Minimize(cp.sum_squares(x))
                constraints = [dmbrs.T @ x == c_mean, x >= 0]

                prob = cp.Problem(objective, constraints)
                prob.solve()

                mbr_indices = np.argsort(x.value)[::-1]
                ind_indices = [i for i in mbr_indices if x.value[i] > 1.0e-6]
                new_basis = dmbrs[ind_indices]

                # And now reduce the new basis if necessary
                new_basis = new_basis[independent_row_indices(new_basis)]

                if len(new_basis) < ph.n_endmembers:
                    logging.info(
                        f"Phase {i_ph} ({ph.name}) is "
                        "rank-deficient ({rank} < {n_mbrs}). "
                        "The transformed solution is described "
                        f"using {len(new_basis)} endmembers."
                    )

                    composite_changed = True

                    # If the composition is set and
                    # consistent with the new basis,
                    # make a new solution with the composition
                    # already set.
                    try:
                        sol = np.linalg.lstsq(
                            new_basis.T, ph.molar_fractions, rcond=None
                        )
                        if sol[1][0] > 1.0e-10:
                            f = None
                        else:
                            f = sol[0]
                    except AttributeError:
                        f = None
                    new_name = f"{ph.name} (transformed)"
                    soln = transform_solution_to_new_basis(
                        ph, new_basis, solution_name=new_name, molar_fractions=f
                    )
                    for h, name in enumerate(soln.endmember_names):
                        if name == "User-created endmember":
                            new_name = (
                                f"Derived member (occupancies: {soln.endmembers[h][1]})"
                            )
                            soln.endmembers[h][0].name = new_name
                            soln.endmember_names[h] = new_name
                    new_phases.append(soln)
                else:
                    logging.info(
                        "This solution is rank-deficient "
                        f"({rank} < {n_mbrs}), "
                        "but its composition requires all "
                        "independent endmembers."
                    )
            else:
                composite_changed = True
                logging.info(
                    f"Phase {i_ph} ({ph.name}) removed from " "composite (rank = 0)."
                )
        else:
            new_phases.append(ph)

    if composite_changed:
        return Composite(new_phases)
    else:
        return composite



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2024 by the BurnMan team, released under the GNU
# GPL v2 or later.


"""
Introducing BurnMan |version|
=============================

Overview
--------

BurnMan is an open source mineral physics and seismological toolkit written in Python
which can enable a user to calculate (or fit) the physical and chemical properties
of endmember minerals, fluids/melts, solutions, and composite assemblages.

Properties which BurnMan can calculate include:

  - the thermodynamic free energies, allowing phase equilibrium calculations,
    endmember activities, chemical potentials and oxygen (and other) fugacities.

  - entropy, enabling the user to calculate isentropes for a given assemblage.

  - volume, to allow the user to create density profiles.

  - seismic velocities, including Voigt-Reuss-Hill and Hashin-Strikman bounds
    and averages.

The toolkit itself comes with a large set of classes and functions which are
designed to allow the user to easily combine mineral physics with
geophysics, and geodynamics. The features of BurnMan include:

  - the full codebase, which includes implementations of many static and thermal equations of state
    (including Vinet, Birch Murnaghan, Mie-Debye-Grueneisen, Modified Tait),
    and solution models (ideal, symmetric, asymmetric, subregular).
  - popular endmember and solution datasets already coded into burnman-usable format
    (including :cite:`HP2011`, :cite:`Stixrude2005`, :cite:`Stixrude2011` and :cite:`Stixrude2022`)
  - Optimal least squares fitting routines for multivariate data with (potentially correlated) errors
    in pressure and temperature. As an example, such functions can be used to
    simultaneously fit volumes, seismic velocities and enthalpies.
  - a "Planet" class, which self-consistently calculates gravity profiles, mass, moment of
    inertia of planets given the chemical and temperature structure of a planet
  - published geotherms
  - a tutorial on the basic use of BurnMan
  - a large collection of annotated examples
  - a set of high-level functions which create files readable by seismological and geodynamic software,
    including: Mineos :cite:`Masters2011`, AxiSEM :cite:`NissenMeyer2014` and ASPECT
  - an extensive suite of unit tests to ensure code functions as intended
  - a series of benchmarks comparing BurnMan output with published data
  - a directory containing user-contributed code from published papers

BurnMan makes extensive use of `SciPy <http://www.scipy.org/>`_,
`NumPy <http://www.numpy.org/>`_ and `SymPy <http://www.sympy.org/>`_
which are widely used Python libraries for scientific computation.
`Matplotlib <http://matplotlib.org/>`_ is used
to display results and produce publication quality figures.
The computations are consistently formulated in terms of SI units.

The code documentation including class and function descriptions can be found online at
http://burnman.readthedocs.io.

This software has been designed to allow the end-user a great deal of freedom
to do whatever calculations they may wish and to add their own modules.
The underlying Python classes have been designed to make new endmember,
solution and composite models easy to read and create.
We have endeavoured to provide examples and benchmarks which cover the
most popular uses of the software, some of which are included in the figure below.
This list is certainly not exhaustive, and we will definitely have missed interesting
applications. We will be very happy to accept contributions in
form of corrections, examples, or new features.

Structure
---------
.. image:: figures/structure.png


.. _ref-installation:


Requirements
------------

  - Python 3.8+
  - Python modules: NumPy, SciPy, SymPy, Sparse, Matplotlib

Optional modules
^^^^^^^^^^^^^^^^

Needed for some functionality:

  - cvxpy: required for some least squares fitting routines
    and solution polytope calculations.
  - pycddlib: required for solution polytope calculations.
  - autograd: required for esoteric solution models defined using a single excess
    function. Not required for the vast majority of users.

Installation
------------

Installation of BurnMan is mostly platform independent.
As long as you know how to use a terminal, the process should be straightforward.
The following instructions should help, but let us know if you have any problems.

Dependencies
^^^^^^^^^^^^
First, make sure you have a sufficiently recent version of python installed
on your machine (see above for the latest requirements).
To check your version of python, type the following in a terminal:
    python --version
If your version is not recent enough, visit https://www.python.org/ to
find out how to install a newer version.

Once you have checked your version of python, you should make sure you have
installed the python module pip. We will use this module to install BurnMan.
If you don't have it already, you can install it by opening a
terminal window and typing:

    python -m ensurepip --upgrade

Mac users will also need to install Xcode, which can be found in the MacStore.

Stable version
^^^^^^^^^^^^^^

If you are only interested in using BurnMan
(rather than developing the software), and you aren't interested in any of the
latest changes, you can install the stable version by typing
the following into a terminal window:

    python -m pip install burnman

This method of installation does not give you easy access to all the examples,
or the test suite. These can be found in the latest release package which can
be downloaded from https://github.com/geodynamics/burnman/releases.

Development version
^^^^^^^^^^^^^^^^^^^
If you want to install the development version of BurnMan
(with all the latest features), you will first need to download the source code.
The best way to do this is by using git (a version control system).
To install git, follow the instructions at https://git-scm.com/downloads.

Then, using a terminal, navigate to the directory into which you want to
clone the BurnMan repository, and type

    git clone https://github.com/geodynamics/burnman.git

(If you don't want to use git, you can download the current master branch
from https://github.com/geodynamics/burnman/archive/master.zip.)

Once the repository is cloned, navigate to the top-level directory by typing
`cd burnman` in the terminal, and then install BurnMan, either in static mode:
`python -m pip install .` or in development mode
(if you want to develop or modify the code): `python -m pip install -e .`.

Checking that the installation worked
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
To check that the installation has worked, you can run the test suite
(`./test.sh`). This takes a few minutes to run.

A more basic check that BurnMan is installed is to navigate to the Burnman
examples directory and type:

    python example_beginner.py

If figures show up, BurnMan has been installed.


Citing BurnMan
--------------

If you use BurnMan in your work, we ask that you cite the following publications:

  - Myhill, R., Cottaar, S., Heister, T., Rose, I., Unterborn, C.,
    Dannberg, J. and Gassmoeller, R. (2023). BurnMan - a Python toolkit for
    planetary geophysics, geochemistry and thermodynamics. Journal of Open Source Software.
    `(https://doi.org/10.21105/joss.05389) <https://doi.org/10.21105/joss.05389>`_

  - Myhill, R., Cottaar, S., Heister, T., Rose, I., Unterborn, C.,
    Dannberg, J., Gassmoeller, R. and Farla, R. (2024):
    BurnMan v2.1.0 [Software]. Computational Infrastructure for Geodynamics. Zenodo.
    `(https://doi.org/10.5281/zenodo.14238360) <https://doi.org/10.5281/zenodo.14238360>`_

  - Cottaar S., Heister, T., Rose, I., and Unterborn, C., (2014). BurnMan: A
    lower mantle mineral physics toolkit, Geochemistry, Geophysics, and
    Geosystems, 15(4), 1164-1179 `(https://doi.org/10.1002/2013GC005122)
    <https://doi.org/10.1002/2013GC005122>`_

Contributing to BurnMan
-----------------------

If you would like to contribute bug fixes, new functions or new modules
to the existing codebase, please contact us at info@burnman.org or make a
pull request at `https://github.com/geodynamics/burnman <https://github.com/geodynamics/burnman>`_.

BurnMan also includes a contrib directory that contains python and ipython
scripts used to reproduce published results. We welcome the submission of
new contributions to this directory. As with the contribution of code,
please contact us at info@burnman.org or make a pull request at
`https://github.com/geodynamics/burnman <https://github.com/geodynamics/burnman>`_.

Acknowledgement and Support
---------------------------

  - This project was initiated at, and follow-up research support was received
    through, Cooperative Institute of Deep Earth Research, CIDER (NSF FESD
    grant 1135452) -- see `www.deep-earth.org <http://www.deep-earth.org>`_

  - We thank all the members of the CIDER Mg/Si team for their input:
    Valentina Magni, Yu Huang, JiaChao Liu, Marc Hirschmann, and Barbara
    Romanowicz. We also thank Lars Stixrude for providing benchmarking calculations
    and Zack Geballe, Motohiko Murakami, Bill McDonough, Quentin Williams,
    Wendy Panero, and Wolfgang Bangerth for helpful discussions.

  - We thank CIG (`www.geodynamics.org <http://www.geodynamics.org>`_) for support
    and accepting our donation of BurnMan as an official project.

"""
from __future__ import absolute_import
import importlib.metadata

# Low level utility functions
from . import utils
from .utils import geotherm

# Classes and associated functions for representing rocks and minerals:
from .classes.material import Material, material_property
from .classes.perplex import PerplexMaterial
from .classes.mineral import Mineral
from .classes.combinedmineral import CombinedMineral
from .classes.solution import Solution, SolidSolution, RelaxedSolution
from .classes.elasticsolutionmodel import ElasticSolutionModel
from .classes.elasticsolution import ElasticSolution, ElasticSolidSolution
from .classes.composite import Composite
from .classes.calibrant import Calibrant
from .classes.anisotropy import AnisotropicMaterial
from .classes.anisotropicmineral import AnisotropicMineral
from .classes.anisotropicmineral import cell_parameters_to_vectors
from .classes.anisotropicmineral import cell_vectors_to_parameters
from .classes.anisotropicsolution import AnisotropicSolution, RelaxedAnisotropicSolution
from .classes.mineral_helpers import HelperLowHighPressureRockTransition
from .classes.mineral_helpers import HelperSpinTransition
from .classes.mineral_helpers import HelperRockSwitcher

# Other classes
from .classes.composition import Composition
from .classes.layer import Layer, BoundaryLayerPerturbation
from .classes.planet import Planet
from .classes.polytope import MaterialPolytope
from .classes import seismic
from .classes import averaging_schemes

# Mineral library
from . import minerals

# Equations of state
from . import eos

# Calibrants
from . import calibrants

# High level tools
from . import tools
from .tools.equilibration import equilibrate
from .tools.partitioning import calculate_nakajima_fp_pv_partition_coefficient

# Optimization functions
from .optimize import composition_fitting
from .optimize import linear_fitting
from .optimize import nonlinear_fitting
from .optimize import nonlinear_solvers
from .optimize import eos_fitting

__version__ = importlib.metadata.version("burnman")



from __future__ import absolute_import

# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.


import numpy as np
import warnings

from . import modified_tait as mt
from . import murnaghan as murn
from . import equation_of_state as eos

from . import einstein


class HP_TMT(eos.EquationOfState):
    """
    Base class for the thermal equation of state based on
    the generic modified Tait equation of state (class MT),
    as described in :cite:`HP2011`.


    An instance "m" of a Mineral can be assigned this
    equation of state with the command m.set_method('hp_tmt')
    (or by initialising the class with the param
    equation_of_state = 'hp_tmt'
    """

    def volume(self, pressure, temperature, params):
        """
        Returns volume [m^3] as a function of pressure [Pa] and temperature [K]
        EQ 12
        """
        Pth = self.__relative_thermal_pressure(temperature, params)
        return mt.volume(pressure - Pth, params)

    def pressure(self, temperature, volume, params):
        """
        Returns pressure [Pa] as a function of temperature [K] and volume[m^3]
        EQ B7
        """
        Pth = self.__relative_thermal_pressure(temperature, params)
        return mt.modified_tait(params["V_0"] / volume, params) + Pth

    def grueneisen_parameter(self, pressure, temperature, volume, params):
        """
        Returns grueneisen parameter [unitless] as a function of pressure,
        temperature, and volume.
        """
        alpha = self.thermal_expansivity(pressure, temperature, volume, params)
        K_T = self.isothermal_bulk_modulus_reuss(pressure, temperature, volume, params)
        C_V = self.molar_heat_capacity_v(pressure, temperature, volume, params)
        return alpha * K_T * volume / C_V

    def isothermal_bulk_modulus_reuss(self, pressure, temperature, volume, params):
        """
        Returns isothermal bulk modulus [Pa] as a function of pressure [Pa],
        temperature [K], and volume [m^3].  EQ 13+2
        """
        Pth = self.__relative_thermal_pressure(temperature, params)
        return mt.bulk_modulus(pressure - Pth, params)

    # calculate the shear modulus as a function of P, V, and T
    def shear_modulus(self, pressure, temperature, volume, params):
        """
        Not implemented.
        Returns 0.
        Could potentially apply a fixed Poissons ratio as a rough estimate.
        """
        return 0.0

    # Cv, heat capacity at constant volume
    def molar_heat_capacity_v(self, pressure, temperature, volume, params):
        """
        Returns heat capacity at constant volume at the pressure, temperature,
        and volume [J/K/mol].
        """
        C_p = self.molar_heat_capacity_p(pressure, temperature, volume, params)
        V = self.volume(pressure, temperature, params)
        alpha = self.thermal_expansivity(pressure, temperature, volume, params)
        K_T = self.isothermal_bulk_modulus_reuss(pressure, temperature, volume, params)
        return C_p - V * temperature * alpha * alpha * K_T

    def thermal_expansivity(self, pressure, temperature, volume, params):
        """
        Returns thermal expansivity at the pressure, temperature,
        and volume [1/K]. This function replaces -Pth in EQ 13+1
        with P-Pth for non-ambient temperature
        """
        a, b, c = mt.tait_constants(params)
        Pth = self.__relative_thermal_pressure(temperature, params)
        psubpth = pressure - params["P_0"] - Pth

        C_V0 = einstein.molar_heat_capacity_v(
            params["T_0"], params["T_einstein"], params["n"]
        )
        C_V = einstein.molar_heat_capacity_v(
            temperature, params["T_einstein"], params["n"]
        )
        alpha = (
            params["a_0"]
            * (C_V / C_V0)
            * 1.0
            / ((1.0 + b * psubpth) * (a + (1.0 - a) * np.power((1 + b * psubpth), c)))
        )
        return alpha

    def molar_heat_capacity_p0(self, temperature, params):
        """
        Returns heat capacity at ambient pressure as a function of temperature
        [J/K/mol]. Cp = a + bT + cT^-2 + dT^-0.5 in :cite:`HP2011`.
        """
        Cp = (
            params["Cp"][0]
            + params["Cp"][1] * temperature
            + params["Cp"][2] * np.power(temperature, -2.0)
            + params["Cp"][3] * np.power(temperature, -0.5)
        )
        return Cp

    def molar_heat_capacity_p_einstein(self, pressure, temperature, volume, params):
        """
        Returns heat capacity at constant pressure at the pressure,
        temperature, and volume, using the C_v and Einstein model [J/K/mol]
        WARNING: Only for comparison with internally self-consistent C_p
        """
        alpha = self.thermal_expansivity(pressure, temperature, volume, params)
        gr = self.grueneisen_parameter(pressure, temperature, volume, params)
        C_v = self.molar_heat_capacity_v(pressure, temperature, volume, params)
        C_p = C_v * (1.0 + gr * alpha * temperature)
        return C_p

    def isentropic_bulk_modulus_reuss(self, pressure, temperature, volume, params):
        """
        Returns adiabatic bulk modulus [Pa] as a function of pressure [Pa],
        temperature [K], and volume [m^3].
        """
        K_T = self.isothermal_bulk_modulus_reuss(pressure, temperature, volume, params)
        C_p = self.molar_heat_capacity_p(pressure, temperature, volume, params)
        C_v = self.molar_heat_capacity_v(pressure, temperature, volume, params)
        K_S = K_T * C_p / C_v
        return K_S

    def gibbs_free_energy(self, pressure, temperature, volume, params):
        """
        Returns the gibbs free energy [J/mol] as a function of pressure [Pa]
        and temperature [K].
        """
        # Calculate temperature and pressure integrals
        a, b, c = mt.tait_constants(params)
        Pth = self.__relative_thermal_pressure(temperature, params)

        psubpth = pressure - params["P_0"] - Pth

        # EQ 13
        if pressure != params["P_0"]:
            intVdP = (
                (pressure - params["P_0"])
                * params["V_0"]
                * (
                    1.0
                    - a
                    + (
                        a
                        * (
                            np.power((1.0 - b * Pth), 1.0 - c)
                            - np.power((1.0 + b * (psubpth)), 1.0 - c)
                        )
                        / (b * (c - 1.0) * (pressure - params["P_0"]))
                    )
                )
            )
        else:
            intVdP = 0.0
        return (
            params["H_0"]
            + self.__intCpdT(temperature, params)
            - temperature * (params["S_0"] + self.__intCpoverTdT(temperature, params))
            + intVdP
        )

    def helmholtz_free_energy(self, pressure, temperature, volume, params):
        return self.gibbs_free_energy(
            pressure, temperature, volume, params
        ) - pressure * self.volume(pressure, temperature, params)

    def entropy(self, pressure, temperature, volume, params):
        """
        Returns the entropy [J/K/mol] as a function of pressure [Pa]
        and temperature [K].
        """
        a, b, c = mt.tait_constants(params)
        Pth = self.__relative_thermal_pressure(temperature, params)

        ksi_over_ksi_0 = einstein.molar_heat_capacity_v(
            temperature, params["T_einstein"], params["n"]
        ) / einstein.molar_heat_capacity_v(
            params["T_0"], params["T_einstein"], params["n"]
        )

        dintVdpdT = (
            params["V_0"] * params["a_0"] * params["K_0"] * a * ksi_over_ksi_0
        ) * (
            np.power((1.0 + b * (pressure - params["P_0"] - Pth)), 0.0 - c)
            - np.power((1.0 - b * Pth), 0.0 - c)
        )
        return params["S_0"] + self.__intCpoverTdT(temperature, params) + dintVdpdT

    def enthalpy(self, pressure, temperature, volume, params):
        """
        Returns the enthalpy [J/mol] as a function of pressure [Pa]
        and temperature [K].
        """
        gibbs = self.gibbs_free_energy(pressure, temperature, volume, params)
        entropy = self.entropy(pressure, temperature, volume, params)
        return gibbs + temperature * entropy

    def molar_heat_capacity_p(self, pressure, temperature, volume, params):
        """
        Returns the heat capacity [J/K/mol] as a function of pressure [Pa]
        and temperature [K].
        """
        a, b, c = mt.tait_constants(params)
        T = temperature
        T_e = params["T_einstein"]
        n = params["n"]
        Pth = self.__relative_thermal_pressure(T, params)

        ksi_over_ksi_0 = einstein.molar_heat_capacity_v(
            T, T_e, n
        ) / einstein.molar_heat_capacity_v(params["T_0"], T_e, n)

        dintVdpdT = (
            params["V_0"] * params["a_0"] * params["K_0"] * a * ksi_over_ksi_0
        ) * (
            np.power((1.0 + b * (pressure - params["P_0"] - Pth)), 0.0 - c)
            - np.power((1.0 - b * Pth), 0.0 - c)
        )

        dSdT0 = (
            params["V_0"]
            * params["K_0"]
            * np.power((ksi_over_ksi_0 * params["a_0"]), 2.0)
            * (
                np.power((1.0 + b * (pressure - params["P_0"] - Pth)), -1.0 - c)
                - np.power((1.0 + b * (-Pth)), -1.0 - c)
            )
        )

        x = T_e / T
        dCv_einstdT = -(
            einstein.molar_heat_capacity_v(T, T_e, n)
            * (1 - 2.0 / x + 2.0 / (np.exp(x) - 1.0))
            * x
            / T
        )

        dSdT1 = -dintVdpdT * dCv_einstdT / einstein.molar_heat_capacity_v(T, T_e, n)

        dSdT = dSdT0 + dSdT1
        return self.molar_heat_capacity_p0(temperature, params) + temperature * dSdT

    def __thermal_pressure(self, T, params):
        """
        Returns thermal pressure [Pa] as a function of T [K]
        EQ 12 - 1 of :cite:`HP2011`.
        """

        # This is basically the mie-gruneisen equation of state for thermal
        # pressure using an Einstein model for heat capacity.  The additional
        # assumption that they make is that alpha*K/Cv, (or gamma / V) is
        # constant over a wide range of compressions.

        # Note that the xi function in HP2011 is just the Einstein heat capacity
        # divided by 3nR. This function is *not* used to calculate the
        # heat capacity - Holland and Powell (2011) prefer the additional
        # freedom provided by their polynomial expression.

        E_th = einstein.thermal_energy(T, params["T_einstein"], params["n"])
        C_V0 = einstein.molar_heat_capacity_v(
            params["T_0"], params["T_einstein"], params["n"]
        )
        P_th = params["a_0"] * params["K_0"] / C_V0 * E_th
        return P_th

    def __relative_thermal_pressure(self, T, params):
        """
        Returns relative thermal pressure [Pa] as a function of T-params['T_0'] [K]
        EQ 12 - 1 of :cite:`HP2011`.
        """
        return self.__thermal_pressure(T, params) - self.__thermal_pressure(
            params["T_0"], params
        )

    def __intCpdT(self, temperature, params):
        """
        Returns the thermal addition to the standard state enthalpy [J/mol]
        at ambient pressure [Pa]
        """
        return (
            params["Cp"][0] * temperature
            + 0.5 * params["Cp"][1] * np.power(temperature, 2.0)
            - params["Cp"][2] / temperature
            + 2.0 * params["Cp"][3] * np.sqrt(temperature)
        ) - (
            params["Cp"][0] * params["T_0"]
            + 0.5 * params["Cp"][1] * params["T_0"] * params["T_0"]
            - params["Cp"][2] / params["T_0"]
            + 2.0 * params["Cp"][3] * np.sqrt(params["T_0"])
        )

    def __intCpoverTdT(self, temperature, params):
        """
        Returns the thermal addition to the standard state entropy [J/K/mol]
        at ambient pressure [Pa]
        """
        return (
            params["Cp"][0] * np.log(temperature)
            + params["Cp"][1] * temperature
            - 0.5 * params["Cp"][2] / np.power(temperature, 2.0)
            - 2.0 * params["Cp"][3] / np.sqrt(temperature)
        ) - (
            params["Cp"][0] * np.log(params["T_0"])
            + params["Cp"][1] * params["T_0"]
            - 0.5 * params["Cp"][2] / (params["T_0"] * params["T_0"])
            - 2.0 * params["Cp"][3] / np.sqrt(params["T_0"])
        )

    def validate_parameters(self, params):
        """
        Check for existence and validity of the parameters
        """
        if "T_0" not in params:
            params["T_0"] = 298.15

        # If standard state enthalpy and entropy are not included
        # this is presumably deliberate, as we can model density
        # and bulk modulus just fine without them.
        # Just add them to the dictionary as nans.
        if "H_0" not in params:
            params["H_0"] = float("nan")
        if "S_0" not in params:
            params["S_0"] = float("nan")

        # First, let's check the EoS parameters for Tref
        mt.MT.validate_parameters(mt.MT(), params)

        # Now check all the required keys for the
        # thermal part of the EoS are in the dictionary
        expected_keys = ["H_0", "S_0", "V_0", "Cp", "a_0", "n", "molar_mass"]
        for k in expected_keys:
            if k not in params:
                raise KeyError("params object missing parameter : " + k)

        # The following line estimates the Einstein temperature
        # according to the empirical equation of
        # Holland and Powell, 2011; base of p.346, para.1
        if "T_einstein" not in params:
            params["T_einstein"] = 10636.0 / (params["S_0"] / params["n"] + 6.44)

        # Finally, check that the values are reasonable.
        if params["T_0"] < 0.0:
            warnings.warn("Unusual value for T_0", stacklevel=2)
        if params["G_0"] is not float("nan") and (
            params["G_0"] < 0.0 or params["G_0"] > 1.0e13
        ):
            warnings.warn("Unusual value for G_0", stacklevel=2)
        if params["Gprime_0"] is not float("nan") and (
            params["Gprime_0"] < -5.0 or params["Gprime_0"] > 10.0
        ):
            warnings.warn("Unusual value for Gprime_0", stacklevel=2)

        # no test for H_0
        if params["S_0"] is not float("nan") and params["S_0"] < 0.0:
            warnings.warn("Unusual value for S_0", stacklevel=2)
        if params["V_0"] < 1.0e-7 or params["V_0"] > 1.0e-2:
            warnings.warn("Unusual value for V_0", stacklevel=2)

        if self.molar_heat_capacity_p0(params["T_0"], params) < 0.0:
            warnings.warn("Negative heat capacity at T_0", stacklevel=2)
        if self.molar_heat_capacity_p0(2000.0, params) < 0.0:
            warnings.warn("Negative heat capacity at 2000K", stacklevel=2)

        if params["a_0"] < 0.0 or params["a_0"] > 1.0e-3:
            warnings.warn("Unusual value for a_0", stacklevel=2)

        if params["n"] < 1.0 or params["n"] > 1000.0:
            warnings.warn("Unusual value for n", stacklevel=2)
        if params["molar_mass"] < 0.001 or params["molar_mass"] > 10.0:
            warnings.warn("Unusual value for molar_mass", stacklevel=2)


class HP_TMTL(eos.EquationOfState):
    """
    Base class for the thermal equation of state
    described in :cite:`HP1998`, but with the Modified Tait as the static part,
    as described in :cite:`HP2011`.

    An instance "m" of a Mineral can be assigned this
    equation of state with the command m.set_method('hp_tmtL')
    (or by initialising the class with the param
    equation_of_state = 'hp_tmtL'
    """

    def _V_T_1bar(self, temperature, params):
        # Constant thermal expansivity at standard state pressure
        # (p.348 of HP2011)
        return params["V_0"] * np.exp(params["a_0"] * (temperature - params["T_0"]))

    def _K_T_1bar(self, temperature, params):
        # Linear bulk modulus dependence as in HP1998 (p.348 of HP2011)
        return params["K_0"] + params["dKdT_0"] * (temperature - params["T_0"])

    def volume(self, pressure, temperature, params):
        """
        Returns volume [m^3] as a function of pressure [Pa] and temperature [K]
        """
        self.static_params["V_0"] = self._V_T_1bar(temperature, params)
        self.static_params["K_0"] = self._K_T_1bar(temperature, params)
        return mt.volume(pressure, self.static_params)

    def pressure(self, temperature, volume, params):
        """
        Returns pressure [Pa] as a function of temperature [K] and volume[m^3]
        """
        self.static_params["V_0"] = self._V_T_1bar(temperature, params)
        self.static_params["K_0"] = self._K_T_1bar(temperature, params)
        return mt.pressure(volume, self.static_params)

    def grueneisen_parameter(self, pressure, temperature, volume, params):
        """
        Returns grueneisen parameter [unitless] as a function of pressure,
        temperature, and volume.
        """
        alpha = self.thermal_expansivity(pressure, temperature, volume, params)
        K_T = self.isothermal_bulk_modulus_reuss(pressure, temperature, volume, params)
        C_V = self.molar_heat_capacity_v(pressure, temperature, volume, params)
        return alpha * K_T * volume / C_V

    def isothermal_bulk_modulus_reuss(self, pressure, temperature, volume, params):
        """
        Returns isothermal bulk modulus [Pa] as a function of pressure [Pa],
        temperature [K], and volume [m^3].
        """
        self.static_params["V_0"] = self._V_T_1bar(temperature, params)
        self.static_params["K_0"] = self._K_T_1bar(temperature, params)
        return mt.bulk_modulus(pressure, self.static_params)

    # calculate the shear modulus as a function of P, V, and T
    def shear_modulus(self, pressure, temperature, volume, params):
        """
        Not implemented.
        Returns 0.
        Could potentially apply a fixed Poissons ratio as a rough estimate.
        """
        return 0.0

    # Cv, heat capacity at constant volume
    def molar_heat_capacity_v(self, pressure, temperature, volume, params):
        """
        Returns heat capacity at constant volume at the pressure, temperature,
        and volume [J/K/mol].
        """
        C_p = self.molar_heat_capacity_p(pressure, temperature, volume, params)
        V = self.volume(pressure, temperature, params)
        alpha = self.thermal_expansivity(pressure, temperature, volume, params)
        K_T = self.isothermal_bulk_modulus_reuss(pressure, temperature, volume, params)
        return C_p - V * temperature * alpha * alpha * K_T

    def thermal_expansivity(self, pressure, temperature, volume, params):
        """
        Returns thermal expansivity at the pressure, temperature,
        and volume [1/K]
        """
        # The derivation of the high pressure thermal expansivity is tedious,
        # so here we take a numerical derivative.
        # TODO Derive and use the analytical derivative.
        dT = 0.1
        self.static_params["V_0"] = self._V_T_1bar(temperature + dT / 2.0, params)
        self.static_params["K_0"] = self._K_T_1bar(temperature + dT / 2.0, params)
        volume1 = mt.volume(pressure, self.static_params)
        self.static_params["V_0"] = self._V_T_1bar(temperature - dT / 2.0, params)
        self.static_params["K_0"] = self._K_T_1bar(temperature - dT / 2.0, params)
        volume0 = mt.volume(pressure, self.static_params)

        return 2.0 * (volume1 - volume0) / (volume1 + volume0) / dT

    def molar_heat_capacity_p0(self, temperature, params):
        """
        Returns heat capacity at ambient pressure as a function of temperature
        [J/K/mol]
        Cp = a + bT + cT^-2 + dT^-0.5 in :cite:`HP1998`.
        """
        Cp = (
            params["Cp"][0]
            + params["Cp"][1] * temperature
            + params["Cp"][2] * np.power(temperature, -2.0)
            + params["Cp"][3] * np.power(temperature, -0.5)
        )
        return Cp

    def isentropic_bulk_modulus_reuss(self, pressure, temperature, volume, params):
        """
        Returns adiabatic bulk modulus [Pa] as a function of pressure [Pa],
        temperature [K], and volume [m^3].
        """
        K_T = self.isothermal_bulk_modulus_reuss(pressure, temperature, volume, params)
        C_p = self.molar_heat_capacity_p(pressure, temperature, volume, params)
        C_v = self.molar_heat_capacity_v(pressure, temperature, volume, params)
        K_S = K_T * C_p / C_v
        return K_S

    def gibbs_free_energy(self, pressure, temperature, volume, params):
        """
        Returns the gibbs free energy [J/mol] as a function of pressure [Pa]
        and temperature [K].
        """
        self.static_params["V_0"] = self._V_T_1bar(temperature, params)
        self.static_params["K_0"] = self._K_T_1bar(temperature, params)
        return (
            params["H_0"]
            + self.__intCpdT(temperature, params)
            - temperature * (params["S_0"] + self.__intCpoverTdT(temperature, params))
            + mt.intVdP(pressure, self.static_params)
        )

    def helmholtz_free_energy(self, pressure, temperature, volume, params):
        return self.gibbs_free_energy(
            pressure, temperature, volume, params
        ) - pressure * self.volume(pressure, temperature, params)

    def entropy(self, pressure, temperature, volume, params):
        """
        Returns the entropy [J/K/mol] as a function of pressure [Pa]
        and temperature [K].
        """
        # The derivation of the entropy is tedious,
        # so here we take a numerical derivative.
        # TODO Derive and use the analytical derivative.
        dT = 0.1
        G1 = self.gibbs_free_energy(pressure, temperature + dT / 2.0, volume, params)
        G0 = self.gibbs_free_energy(pressure, temperature - dT / 2.0, volume, params)

        return (G0 - G1) / dT

    def enthalpy(self, pressure, temperature, volume, params):
        """
        Returns the enthalpy [J/mol] as a function of pressure [Pa]
        and temperature [K].
        """
        gibbs = self.gibbs_free_energy(pressure, temperature, volume, params)
        entropy = self.entropy(pressure, temperature, volume, params)
        return gibbs + temperature * entropy

    def molar_heat_capacity_p(self, pressure, temperature, volume, params):
        """
        Returns the heat capacity [J/K/mol] as a function of pressure [Pa]
        and temperature [K].
        """
        # The differentiation is tedious, so for now we just take the
        # numerical derivative of S
        # TODO Derive and use the analytical derivative.
        dT = 0.1
        S1 = self.entropy(pressure, temperature + dT / 2.0, volume, params)
        S0 = self.entropy(pressure, temperature - dT / 2.0, volume, params)
        return temperature * (S1 - S0) / dT

    def __intCpdT(self, temperature, params):
        """
        Returns the thermal addition to the standard state enthalpy [J/mol]
        at ambient pressure [Pa]
        """
        return (
            params["Cp"][0] * temperature
            + 0.5 * params["Cp"][1] * np.power(temperature, 2.0)
            - params["Cp"][2] / temperature
            + 2.0 * params["Cp"][3] * np.sqrt(temperature)
        ) - (
            params["Cp"][0] * params["T_0"]
            + 0.5 * params["Cp"][1] * params["T_0"] * params["T_0"]
            - params["Cp"][2] / params["T_0"]
            + 2.0 * params["Cp"][3] * np.sqrt(params["T_0"])
        )

    def __intCpoverTdT(self, temperature, params):
        """
        Returns the thermal addition to the standard state entropy [J/K/mol]
        at ambient pressure [Pa]
        """
        return (
            params["Cp"][0] * np.log(temperature)
            + params["Cp"][1] * temperature
            - 0.5 * params["Cp"][2] / np.power(temperature, 2.0)
            - 2.0 * params["Cp"][3] / np.sqrt(temperature)
        ) - (
            params["Cp"][0] * np.log(params["T_0"])
            + params["Cp"][1] * params["T_0"]
            - 0.5 * params["Cp"][2] / (params["T_0"] * params["T_0"])
            - 2.0 * params["Cp"][3] / np.sqrt(params["T_0"])
        )

    def validate_parameters(self, params):
        """
        Check for existence and validity of the parameters
        """
        if "T_0" not in params:
            params["T_0"] = 298.15

        # If standard state enthalpy and entropy are not included
        # this is presumably deliberate, as we can model density
        # and bulk modulus just fine without them.
        # Just add them to the dictionary as nans.
        if "H_0" not in params:
            params["H_0"] = float("nan")
        if "S_0" not in params:
            params["S_0"] = float("nan")

        # First, let's check the EoS parameters for Tref
        mt.MT.validate_parameters(mt.MT(), params)
        self.static_params = {
            "V_0": params["V_0"],
            "K_0": params["K_0"],
            "Kprime_0": params["Kprime_0"],
            "Kdprime_0": params["Kdprime_0"],
            "P_0": params["P_0"],
        }

        # Now check all the required keys for the
        # thermal part of the EoS are in the dictionary
        expected_keys = ["H_0", "S_0", "V_0", "Cp", "a_0", "dKdT_0", "n", "molar_mass"]
        for k in expected_keys:
            if k not in params:
                raise KeyError("params object missing parameter : " + k)

        # Finally, check that the values are reasonable.
        if params["T_0"] < 0.0:
            warnings.warn("Unusual value for T_0", stacklevel=2)
        if params["G_0"] is not float("nan") and (
            params["G_0"] < 0.0 or params["G_0"] > 1.0e13
        ):
            warnings.warn("Unusual value for G_0", stacklevel=2)
        if params["Gprime_0"] is not float("nan") and (
            params["Gprime_0"] < -5.0 or params["Gprime_0"] > 10.0
        ):
            warnings.warn("Unusual value for Gprime_0", stacklevel=2)

        # no test for H_0 or S_0 (several HP endmembers have S_0 < 0)
        if params["V_0"] < 1.0e-7 or params["V_0"] > 1.0e-2:
            warnings.warn("Unusual value for V_0", stacklevel=2)

        if self.molar_heat_capacity_p0(params["T_0"], params) < 0.0:
            warnings.warn("Negative heat capacity at T_0", stacklevel=2)
        if self.molar_heat_capacity_p0(2000.0, params) < 0.0:
            warnings.warn("Negative heat capacity at 2000K", stacklevel=2)

        if params["a_0"] < 0.0 or params["a_0"] > 1.0e-3:
            warnings.warn("Unusual value for a_0", stacklevel=2)

        if params["n"] < 1.0 or params["n"] > 1000.0:
            warnings.warn("Unusual value for n", stacklevel=2)
        if params["molar_mass"] < 0.001 or params["molar_mass"] > 10.0:
            warnings.warn("Unusual value for molar_mass", stacklevel=2)


class HP98(eos.EquationOfState):
    """
    Base class for the thermal equation of state
    described in :cite:`HP1998`.

    An instance "m" of a Mineral can be assigned this
    equation of state with the command m.set_method('hp98')
    (or by initialising the class with the param
    equation_of_state = 'hp98'
    """

    def _V_T_1bar(self, temperature, params):
        return params["V_0"] * (
            1.0
            + params["a_0"] * (temperature - params["T_0"])
            - 20.0 * params["a_0"] * (np.sqrt(temperature) - np.sqrt(params["T_0"]))
        )

    def _K_T_1bar(self, temperature, params):
        return params["K_0"] + params["dKdT_0"] * (temperature - params["T_0"])

    def volume(self, pressure, temperature, params):
        """
        Returns volume [m^3] as a function of pressure [Pa] and temperature [K]
        """
        return murn.volume(
            pressure,
            self._V_T_1bar(temperature, params),
            self._K_T_1bar(temperature, params),
            params["Kprime_0"],
        )

    def pressure(self, temperature, volume, params):
        """
        Returns pressure [Pa] as a function of temperature [K] and volume[m^3]
        """
        return murn.pressure(
            volume,
            self._V_T_1bar(temperature, params),
            self._K_T_1bar(temperature, params),
            params["Kprime_0"],
        )

    def grueneisen_parameter(self, pressure, temperature, volume, params):
        """
        Returns grueneisen parameter [unitless] as a function of pressure,
        temperature, and volume.
        """
        alpha = self.thermal_expansivity(pressure, temperature, volume, params)
        K_T = self.isothermal_bulk_modulus_reuss(pressure, temperature, volume, params)
        C_V = self.molar_heat_capacity_v(pressure, temperature, volume, params)
        return alpha * K_T * volume / C_V

    def isothermal_bulk_modulus_reuss(self, pressure, temperature, volume, params):
        """
        Returns isothermal bulk modulus [Pa] as a function of pressure [Pa],
        temperature [K], and volume [m^3].
        """
        return murn.bulk_modulus(
            pressure, self._K_T_1bar(temperature, params), params["Kprime_0"]
        )

    # calculate the shear modulus as a function of P, V, and T
    def shear_modulus(self, pressure, temperature, volume, params):
        """
        Not implemented.
        Returns 0.
        Could potentially apply a fixed Poissons ratio as a rough estimate.
        """
        return 0.0

    # Cv, heat capacity at constant volume
    def molar_heat_capacity_v(self, pressure, temperature, volume, params):
        """
        Returns heat capacity at constant volume at the pressure, temperature,
        and volume [J/K/mol].
        """
        C_p = self.molar_heat_capacity_p(pressure, temperature, volume, params)
        V = self.volume(pressure, temperature, params)
        alpha = self.thermal_expansivity(pressure, temperature, volume, params)
        K_T = self.isothermal_bulk_modulus_reuss(pressure, temperature, volume, params)
        return C_p - V * temperature * alpha * alpha * K_T

    def thermal_expansivity(self, pressure, temperature, volume, params):
        """
        Returns thermal expansivity at the pressure, temperature,
        and volume [1/K]
        """
        VT = self._V_T_1bar(temperature, params)
        KT = self._K_T_1bar(temperature, params)
        volume = murn.volume(pressure, VT, KT, params["Kprime_0"])
        dVTdT = params["V_0"] * params["a_0"] * (1.0 - 10.0 / np.sqrt(temperature))
        g = volume / VT
        dgdKT = (
            pressure
            * np.power(
                1.0 + pressure * params["Kprime_0"] / KT, -1 - 1.0 / params["Kprime_0"]
            )
            / (KT * KT)
        )
        dVdT = dVTdT * g + VT * dgdKT * params["dKdT_0"]
        return dVdT / volume

    def molar_heat_capacity_p0(self, temperature, params):
        """
        Returns heat capacity at ambient pressure as a function of temperature
        [J/K/mol]
        Cp = a + bT + cT^-2 + dT^-0.5 in :cite:`HP1998`.
        """
        Cp = (
            params["Cp"][0]
            + params["Cp"][1] * temperature
            + params["Cp"][2] * np.power(temperature, -2.0)
            + params["Cp"][3] * np.power(temperature, -0.5)
        )
        return Cp

    def isentropic_bulk_modulus_reuss(self, pressure, temperature, volume, params):
        """
        Returns adiabatic bulk modulus [Pa] as a function of pressure [Pa],
        temperature [K], and volume [m^3].
        """
        K_T = self.isothermal_bulk_modulus_reuss(pressure, temperature, volume, params)
        C_p = self.molar_heat_capacity_p(pressure, temperature, volume, params)
        C_v = self.molar_heat_capacity_v(pressure, temperature, volume, params)
        K_S = K_T * C_p / C_v
        return K_S

    def gibbs_free_energy(self, pressure, temperature, volume, params):
        """
        Returns the gibbs free energy [J/mol] as a function of pressure [Pa]
        and temperature [K].
        """
        return (
            params["H_0"]
            + self.__intCpdT(temperature, params)
            - temperature * (params["S_0"] + self.__intCpoverTdT(temperature, params))
            + murn.intVdP(
                pressure,
                self._V_T_1bar(temperature, params),
                self._K_T_1bar(temperature, params),
                params["Kprime_0"],
            )
        )

    def helmholtz_free_energy(self, pressure, temperature, volume, params):
        return self.gibbs_free_energy(
            pressure, temperature, volume, params
        ) - pressure * self.volume(pressure, temperature, params)

    def entropy(self, pressure, temperature, volume, params):
        """
        Returns the entropy [J/K/mol] as a function of pressure [Pa]
        and temperature [K].
        """
        # The entropy involves differentiating intdVdp
        # with respect to temperature.
        # We do this using the product and chain rules

        VT = self._V_T_1bar(temperature, params)
        KT = self._K_T_1bar(temperature, params)
        dVTdT = params["V_0"] * params["a_0"] * (1.0 - 10.0 / np.sqrt(temperature))
        g = murn.intVdP(pressure, VT, KT, params["Kprime_0"]) / VT
        dgdKT = (
            (pressure / KT + 1.0)
            * np.power(
                1.0 + pressure * params["Kprime_0"] / KT, -1.0 / params["Kprime_0"]
            )
            - 1.0
        ) / (params["Kprime_0"] - 1.0)
        dintVdpdT = dVTdT * g + VT * dgdKT * params["dKdT_0"]
        return params["S_0"] + self.__intCpoverTdT(temperature, params) - dintVdpdT

    def enthalpy(self, pressure, temperature, volume, params):
        """
        Returns the enthalpy [J/mol] as a function of pressure [Pa]
        and temperature [K].
        """
        gibbs = self.gibbs_free_energy(pressure, temperature, volume, params)
        entropy = self.entropy(pressure, temperature, volume, params)
        return gibbs + temperature * entropy

    def molar_heat_capacity_p(self, pressure, temperature, volume, params):
        """
        Returns the heat capacity [J/K/mol] as a function of pressure [Pa]
        and temperature [K].
        """
        # The differentiation is tedious, so for now we just take the
        # numerical derivative of S
        # TODO calculate the analytical derivative
        dT = 0.1
        S1 = self.entropy(pressure, temperature + dT / 2.0, volume, params)
        S0 = self.entropy(pressure, temperature - dT / 2.0, volume, params)
        return temperature * (S1 - S0) / dT

    def __intCpdT(self, temperature, params):
        """
        Returns the thermal addition to the standard state enthalpy [J/mol]
        at ambient pressure [Pa]
        """
        return (
            params["Cp"][0] * temperature
            + 0.5 * params["Cp"][1] * np.power(temperature, 2.0)
            - params["Cp"][2] / temperature
            + 2.0 * params["Cp"][3] * np.sqrt(temperature)
        ) - (
            params["Cp"][0] * params["T_0"]
            + 0.5 * params["Cp"][1] * params["T_0"] * params["T_0"]
            - params["Cp"][2] / params["T_0"]
            + 2.0 * params["Cp"][3] * np.sqrt(params["T_0"])
        )

    def __intCpoverTdT(self, temperature, params):
        """
        Returns the thermal addition to the standard state entropy [J/K/mol]
        at ambient pressure [Pa]
        """
        return (
            params["Cp"][0] * np.log(temperature)
            + params["Cp"][1] * temperature
            - 0.5 * params["Cp"][2] / np.power(temperature, 2.0)
            - 2.0 * params["Cp"][3] / np.sqrt(temperature)
        ) - (
            params["Cp"][0] * np.log(params["T_0"])
            + params["Cp"][1] * params["T_0"]
            - 0.5 * params["Cp"][2] / (params["T_0"] * params["T_0"])
            - 2.0 * params["Cp"][3] / np.sqrt(params["T_0"])
        )

    def validate_parameters(self, params):
        """
        Check for existence and validity of the parameters
        """
        if "T_0" not in params:
            params["T_0"] = 298.15

        # If standard state enthalpy and entropy are not included
        # this is presumably deliberate, as we can model density
        # and bulk modulus just fine without them.
        # Just add them to the dictionary as nans.
        if "H_0" not in params:
            params["H_0"] = float("nan")
        if "S_0" not in params:
            params["S_0"] = float("nan")

        # First, let's check the EoS parameters for Tref
        murn.Murnaghan.validate_parameters(murn.Murnaghan(), params)

        # Now check all the required keys for the
        # thermal part of the EoS are in the dictionary
        expected_keys = ["H_0", "S_0", "V_0", "Cp", "a_0", "dKdT_0", "n", "molar_mass"]
        for k in expected_keys:
            if k not in params:
                raise KeyError("params object missing parameter : " + k)

        # Finally, check that the values are reasonable.
        if params["T_0"] < 0.0:
            warnings.warn("Unusual value for T_0", stacklevel=2)
        if params["G_0"] is not float("nan") and (
            params["G_0"] < 0.0 or params["G_0"] > 1.0e13
        ):
            warnings.warn("Unusual value for G_0", stacklevel=2)
        if params["Gprime_0"] is not float("nan") and (
            params["Gprime_0"] < -5.0 or params["Gprime_0"] > 10.0
        ):
            warnings.warn("Unusual value for Gprime_0", stacklevel=2)

        # no test for H_0 or S_0 (several HP endmembers have S_0 < 0)
        if params["V_0"] < 1.0e-7 or params["V_0"] > 1.0e-2:
            warnings.warn("Unusual value for V_0", stacklevel=2)

        if self.molar_heat_capacity_p0(params["T_0"], params) < 0.0:
            warnings.warn("Negative heat capacity at T_0", stacklevel=2)
        if self.molar_heat_capacity_p0(2000.0, params) < 0.0:
            warnings.warn("Negative heat capacity at 2000K", stacklevel=2)

        if params["a_0"] < 0.0 or params["a_0"] > 1.0e-3:
            warnings.warn("Unusual value for a_0", stacklevel=2)

        if params["n"] < 1.0 or params["n"] > 1000.0:
            warnings.warn("Unusual value for n", stacklevel=2)
        if params["molar_mass"] < 0.001 or params["molar_mass"] > 10.0:
            warnings.warn("Unusual value for molar_mass", stacklevel=2)



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2024 by the BurnMan team, released under the GNU
# GPL v2 or later.

from burnman.eos.vinet import Vinet
from burnman.eos.birch_murnaghan import BirchMurnaghanBase as BM3
from burnman.eos.mie_grueneisen_debye import MGDBase
from burnman.classes.calibrant import Calibrant
from burnman.utils.unitcell import molar_volume_from_unit_cell_volume

"""
Fei_2007
^^^^^^^^
"""


class Pt(Calibrant):
    """
    The Pt pressure standard reported by
    Fei et al. (2007; https://doi.org/10.1073/pnas.0609013104).
    """

    def __init__(self):
        def _pressure_Fei_Pt(volume, temperature, params):

            # Isothermal pressure (GPa)
            pressure_model = Vinet()
            P0 = pressure_model.pressure(params["T_0"], volume, params)

            # Thermal pressure
            thermal_model = MGDBase()
            Pth0 = thermal_model._thermal_pressure(params["T_0"], volume, params)
            Pth = thermal_model._thermal_pressure(temperature, volume, params)

            # Total pressure
            P = P0 + Pth - Pth0

            return P

        Z = 4.0
        _params_Fei_Pt = {
            "V_0": molar_volume_from_unit_cell_volume(60.38, Z),
            "K_0": 277.0e9,
            "Kprime_0": 5.08,
            "Debye_0": 230.0,
            "grueneisen_0": 2.72,
            "q_0": 0.5,
            "n": 1.0,
            "T_0": 300.0,
            "P_0": 0.0,
            "Z": Z,
        }

        Calibrant.__init__(self, _pressure_Fei_Pt, "pressure", _params_Fei_Pt)


class Au(Calibrant):
    """
    The Au pressure standard reported by
    Fei et al. (2007; https://doi.org/10.1073/pnas.0609013104).
    """

    def __init__(self):
        def _pressure_Fei_Au(volume, temperature, params):

            # Isothermal pressure (GPa)
            pressure_model = Vinet()
            P0 = pressure_model.pressure(params["T_0"], volume, params)

            # Thermal pressure
            thermal_model = MGDBase()
            Pth0 = thermal_model._thermal_pressure(params["T_0"], volume, params)
            Pth = thermal_model._thermal_pressure(temperature, volume, params)

            # Total pressure
            P = P0 + Pth - Pth0

            return P

        Z = 4.0
        _params_Fei_Au = {
            "V_0": molar_volume_from_unit_cell_volume(67.850, Z),
            "K_0": 167.0e9,
            "Kprime_0": 6.0,
            "Debye_0": 170.0,
            "grueneisen_0": 2.97,
            "q_0": 0.6,
            "n": 1.0,
            "T_0": 300.0,
            "P_0": 0.0,
            "Z": Z,
        }

        Calibrant.__init__(self, _pressure_Fei_Au, "pressure", _params_Fei_Au)



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2024 by the BurnMan team, released under the GNU
# GPL v2 or later.

from burnman.eos.vinet import Vinet
from burnman.eos.mie_grueneisen_debye import MGDBase
from burnman.classes.calibrant import Calibrant
import numpy as np


"""
Dewaele_2013
^^^^^^^^^^^^
"""


class Al2O3_corundum(Calibrant):
    """
    The alpha-Al2O3 (corundum) pressure standard reported by
    Dewaele and Torrent (2013; https://doi.org/10.1103/PhysRevB.88.064107).
    The thermal model is from Shi (2022).
    """

    def __init__(self):
        def _pressure_Dewaele_Al2O3(volume, temperature, params):

            # Isothermal pressure (GPa)
            pressure_model = Vinet()
            P0 = pressure_model.pressure(params["T_0"], volume, params)

            # Thermal pressure
            thermal_model = MGDBase()
            Pth0 = thermal_model._thermal_pressure(params["T_0"], volume, params)
            Pth = thermal_model._thermal_pressure(temperature, volume, params)

            # Total pressure
            P = P0 + Pth - Pth0

            return P

        _params_Dewaele_Al2O3 = {
            "V_0": 2.5594e-5,
            "K_0": 254.1e9,
            "Kprime_0": 4.0,
            "Debye_0": 1100.0,
            "grueneisen_0": 1.32,
            "q_0": 0.8,
            "n": 5.0,
            "T_0": 300.0,
            "P_0": 0.0,
            "Z": 6.0,
        }

        Calibrant.__init__(
            self, _pressure_Dewaele_Al2O3, "pressure", _params_Dewaele_Al2O3
        )



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2019 by the BurnMan team, released under the GNU
# GPL v2 or later.


from __future__ import absolute_import
import numpy as np
from .. import constants


def calculate_nakajima_fp_pv_partition_coefficient(
    pressure, temperature, bulk_composition_mol, initial_distribution_coefficient
):
    """
    Calculate the partitioning of iron between periclase and bridgmanite as given
    by Nakajima et al., 2012.

    :param pressure: Equilibrium pressure [Pa]
    :type pressure: float
    :param temperature: Equilibrium temperature [K]
    :type temperature: float
    :param bulk_composition_mol: Bulk composition [mol].
        Only Mg, Fe, and Si are assumed to explicitly affect the partitioning
        with Al playing an implicit role.
    :type bulk_composition_mol: dict
    :param initial_distribution_coefficient: The distribution coefficient (Kd_0)
        at 25 GPa and 0 K.
    :type initial_distribution_coefficient: float

    :returns: The proportion of Fe in ferropericlase and perovskite, respectively
    :rtype: tuple
    """

    norm = bulk_composition_mol["Mg"] + bulk_composition_mol["Fe"]
    f_FeO = bulk_composition_mol["Fe"] / norm
    f_SiO2 = bulk_composition_mol["Si"] / norm

    Kd_0 = initial_distribution_coefficient
    delV = 2.0e-7  # in m^3/mol, average taken from Nakajima et al 2012, JGR

    # eq 5 Nakajima et al 2012, JGR. Solved for ln(K(P,T,X))
    rs = ((25.0e9 - pressure) * delV / (constants.gas_constant * temperature)) + np.log(
        Kd_0
    )

    # The exchange coefficent at P and T. K(P,T,X) in eq 5 Nakajima et al 2012
    K = np.exp(rs)

    # Solving equation 6 in Nakajima et al., 2012 for X_Fe_fp and X_Fe_pv
    # Solved using the definition of the distribution coefficient
    # to define X_Fe_fp as a function of X_Fe_pv

    num_to_sqrt = (-4.0 * f_FeO * (K - 1.0) * K * f_SiO2) + (
        pow(1.0 + (f_FeO * (K - 1)) + ((K - 1.0) * f_SiO2), 2.0)
    )

    X_Fe_pv = (
        -1.0 + f_FeO - (f_FeO * K) + f_SiO2 - (f_SiO2 * K) + np.sqrt(num_to_sqrt)
    ) / (2.0 * f_SiO2 * (1.0 - K))

    X_Fe_fp = X_Fe_pv / (((1.0 - X_Fe_pv) * K) + X_Fe_pv)

    return (X_Fe_fp, X_Fe_pv)



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2022 by the BurnMan team, released under the GNU
# GPL v2 or later.

import numpy as np
import scipy.optimize as opt
from ..utils.math import bracket


class Calibrant(object):
    """
    The base class for a pressure calibrant material.

    :param calibrant_function: A function that takes either pressure,
        temperature and a params object as arguments, returning the volume,
        or takes volume, temperature and a params object, returning the pressure.
    :type calibrant_function: function

    :param calibrant_function_return_type: The return type of the calibrant function.
        Valid values are 'pressure' or 'volume'.
    :type calibrant_function_return_type: str

    :param params: A dictionary containing the parameters required by the
        calibrant function.
    :type params: dictionary
    """

    def __init__(self, calibrant_function, calibrant_function_return_type, params):
        if calibrant_function_return_type == "pressure":
            self.pressure_function = calibrant_function
            self.volume_function = self._volume_using_pressure_function
        elif calibrant_function_return_type == "volume":
            self.volume_function = calibrant_function
            self.pressure_function = self._pressure_using_volume_function
        else:
            raise Exception(
                "calibrant function return type must either " "be pressure or volume"
            )

        self.calibrant_function = calibrant_function

        self.params = params

    def _volume_using_pressure_function(self, pressure, temperature, params):
        """
        Helper function to compute volume iteratively by Brent's method using
        a function of the form pressure(volume, temperature).
        """

        def func(x):
            return self.pressure_function(x, temperature, params) - pressure

        try:
            sol = bracket(func, params["V_0"], 1.0e-2 * params["V_0"])
        except ValueError:
            raise ValueError(
                "Cannot find a volume, perhaps you are outside "
                "of the range of validity for the equation "
                "of state?"
            )
        return opt.brentq(func, sol[0], sol[1])

    def _pressure_using_volume_function(self, volume, temperature, params):
        """
        Helper function to compute pressure iteratively by Brent's method using
        a function of the form volume(pressure, temperature).
        """

        def func(x):
            (self.volume_function(x, temperature, params) - volume)

        try:
            sol = bracket(func, 0.0, 300.0e9)
        except ValueError:
            raise ValueError(
                "Cannot find a pressure, perhaps you are outside "
                "of the range of validity for the equation "
                "of state?"
            )
        return opt.brentq(func, sol[0], sol[1])

    def pressure(self, volume, temperature, VT_covariance=None):
        """
        Returns the pressure of the calibrant as a function of
        volume, temperature and (optionally) a volume-temperature
        variance-covariance matrix.

        :param volume: The volume of the calibrant [m^3/mol].
        :type volume: float

        :param temperature: The temperature of the calibrant [K].
        :type temperature: float

        :param VT_covariance: The volume-temperature
            variance-covariance matrix [optional].
        :type VT_covariance: 2x2 numpy.array

        :returns: The pressure of the calibrant [Pa] and the
            pressure-volume-temperature variance-covariance matrix
            if PT_covariance is provided.
        :rtype: float, tuple of a float and a numpy.array (3x3)
        """
        if VT_covariance is None:
            return self.pressure_function(volume, temperature, self.params)
        else:
            # Here we take the centered differences
            # We could alternatively use thermodynamic properties
            # but these have not yet been implemented.
            dV = volume / 1.0e7
            dT = 0.01
            PdV0 = self.pressure_function(volume - dV / 2.0, temperature, self.params)
            PdV1 = self.pressure_function(volume + dV / 2.0, temperature, self.params)
            PdT0 = self.pressure_function(volume, temperature - dT / 2.0, self.params)
            PdT1 = self.pressure_function(volume, temperature + dT / 2.0, self.params)
            pressure = (PdV0 + PdV1 + PdT0 + PdT1) / 4.0

            gradPVT = np.zeros((2, 3))
            gradPVT[:, 1:] = np.eye(2)
            gradPVT[:, 0] = [(PdV1 - PdV0) / dV, (PdT1 - PdT0) / dT]
            PVT_covariance = gradPVT.T.dot(VT_covariance).dot(gradPVT)
            return pressure, PVT_covariance

    def volume(self, pressure, temperature, PT_covariance=None):
        """
        Returns the volume of the calibrant as a function of
        pressure, temperature and (optionally) a pressure-temperature
        variance-covariance matrix.

        :param pressure: The pressure of the calibrant [Pa].
        :type pressure: float

        :param temperature: The temperature of the calibrant [K].
        :type temperature: float

        :param PT_covariance: The pressure-temperature
            variance-covariance matrix [optional].
        :type PT_covariance: 2x2 numpy.array

        :returns: The volume of the calibrant [m^3/mol] and
            the volume-pressure-temperature variance-covariance matrix
            if VT_covariance is provided.
        :rtype: float, tuple of a float and a numpy.array (3x3)
        """
        if PT_covariance is None:
            return self.volume_function(pressure, temperature, self.params)
        else:
            # Here we take the centered differences
            # We could alternatively use thermodynamic properties
            # but these have not yet been implemented.
            dP = 100.0
            dT = 0.01
            VdP0 = self.volume_function(pressure - dP / 2.0, temperature, self.params)
            VdP1 = self.volume_function(pressure + dP / 2.0, temperature, self.params)
            VdT0 = self.volume_function(pressure, temperature - dT / 2.0, self.params)
            VdT1 = self.volume_function(pressure, temperature + dT / 2.0, self.params)
            volume = (VdP0 + VdP1 + VdT0 + VdT1) / 4.0

            gradVPT = np.zeros((2, 3))
            gradVPT[:, 1:] = np.eye(2)
            gradVPT[:, 0] = [(VdP1 - VdP0) / dP, (VdT1 - VdT0) / dT]
            VPT_covariance = gradVPT.T.dot(PT_covariance).dot(gradVPT)
            return volume, VPT_covariance



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2024 by the BurnMan team, released under the GNU
# GPL v2 or later.

from burnman.eos.vinet import Vinet
from burnman.eos.mie_grueneisen_debye import MGDBase
from burnman.classes.calibrant import Calibrant


"""
Litasov_2013
^^^^^^^^^^^^
"""


class Mo_bcc(Calibrant):
    """
    The BCC Mo pressure standard reported by
    Litasov et al. (2013; https://doi.org/10.1063/1.4794127).
    """

    def __init__(self):
        def _pressure_Litasov_Mo(volume, temperature, params):

            # Isothermal pressure (GPa)
            pressure_model = Vinet()
            P0 = pressure_model.pressure(params["T_0"], volume, params)

            # Thermal pressure
            thermal_model = MGDBase()
            Pth0 = thermal_model._thermal_pressure(params["T_0"], volume, params)
            Pth = thermal_model._thermal_pressure(temperature, volume, params)

            # Total pressure
            P = P0 + Pth - Pth0

            return P

        _params_Litasov_Mo = {
            "V_0": 9.37647e-06,
            "K_0": 260.0e9,
            "Kprime_0": 4.21,
            "Debye_0": 470.0,
            "grueneisen_0": 2.03,
            "q_0": 0.24,
            "n": 1.0,
            "T_0": 300.0,
            "P_0": 0.0,
            "Z": 2.0,
        }

        Calibrant.__init__(self, _pressure_Litasov_Mo, "pressure", _params_Litasov_Mo)


class W_bcc(Calibrant):
    """
    The BCC W pressure standard reported by
    Litasov et al. (2013; https://doi.org/10.1063/1.4799018).
    """

    def __init__(self):
        def _pressure_Litasov_W(volume, temperature, params):

            # Isothermal pressure (GPa)
            pressure_model = Vinet()
            P0 = pressure_model.pressure(params["T_0"], volume, params)

            # Thermal pressure
            thermal_model = MGDBase()
            Pth0 = thermal_model._thermal_pressure(params["T_0"], volume, params)
            Pth = thermal_model._thermal_pressure(temperature, volume, params)

            # Total pressure
            P = P0 + Pth - Pth0

            return P

        _params_Litasov_W = {
            "V_0": 9.5481e-06,
            "K_0": 308.0e9,
            "Kprime_0": 4.2,
            "Debye_0": 388.0,
            "grueneisen_0": 1.81,
            "q_0": 0.71,
            "n": 1.0,
            "T_0": 300.0,
            "P_0": 0.0,
            "Z": 2.0,
        }

        Calibrant.__init__(self, _pressure_Litasov_W, "pressure", _params_Litasov_W)



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.

from __future__ import absolute_import

import numpy as np
from .. import constants

# Try to import the jit from numba.  If it is
# not available, just go with the standard
# python interpreter
try:
    import os

    if "NUMBA_DISABLE_JIT" in os.environ and int(os.environ["NUMBA_DISABLE_JIT"]) == 1:
        raise ImportError("NOOOO!")
    from numba import jit
except ImportError:

    def jit(fn):
        return fn


"""
Functions for the Einstein model of a solid.
"""

eps = np.finfo(float).eps


@jit(nopython=True)
def thermal_energy(T, einstein_T, n):
    """
    calculate the thermal energy of a substance.  Takes the temperature,
    the Einstein temperature, and n, the number of atoms per molecule.
    Returns thermal energy in J/mol
    """
    if T <= eps:
        return 0.0
    x = einstein_T / T
    E_th = 3.0 * n * constants.gas_constant * einstein_T * (1.0 / (np.exp(x) - 1.0))
    return E_th


@jit(nopython=True)
def molar_heat_capacity_v(T, einstein_T, n):
    """
    Heat capacity at constant volume.  In J/K/mol
    """
    if T <= eps:
        return 0.0
    x = einstein_T / T
    C_v = (
        3.0
        * n
        * constants.gas_constant
        * (x * x * np.exp(x) / np.power(np.exp(x) - 1.0, 2.0))
    )
    return C_v


@jit(nopython=True)
def helmholtz_free_energy(T, einstein_T, n):
    """
    Helmholtz free energy of lattice vibrations in the Einstein model [J].
    It is important to note that this does NOT include the zero
    point energy for the lattice.  As long as you are
    calculating relative differences in F, this should cancel anyway.
    """
    E = thermal_energy(T, einstein_T, n)
    S = entropy(T, einstein_T, n)
    return E - T * S


@jit(nopython=True)
def entropy(T, einstein_T, n):
    """
    Entropy due to lattice vibrations in the Einstein model [J/K]
    """
    if T <= eps:
        return 0.0
    x = einstein_T / T
    S = (
        3.0
        * n
        * constants.gas_constant
        * (-x * np.exp(-x) / (np.exp(-x) - 1.0) - np.log(1.0 - np.exp(-x)))
    )
    return S


@jit(nopython=True)
def dmolar_heat_capacity_v_dT(T, einstein_T, n):
    """
    First temperature derivative of the heat capacity at constant volume
    according to the Einstein model [J/K^2/mol].
    """
    if T <= eps:
        return 0.0

    x = einstein_T / T
    dCvdT = (
        3.0
        * n
        * constants.gas_constant
        * x
        * x
        * np.exp(x)
        * ((x - 2.0) * np.exp(x) + (x + 2.0))
        / (T * np.power(np.exp(x) - 1.0, 3.0))
    )

    return dCvdT



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.

from __future__ import absolute_import
from __future__ import print_function
import numpy as np
from itertools import product
from scipy.linalg import lu_factor, lu_solve
from collections import namedtuple

from ..optimize.nonlinear_solvers import damped_newton_solve
from ..classes.solution import Solution


def calculate_constraints(assemblage, n_free_compositional_vectors):
    """
    This function calculates the linear inequality constraints bounding
    the valid parameter space for a given assemblage.

    The constraints are as follows:

      - Pressure and temperature must be positive
      - All phase fractions must be positive
      - All site-species occupancies must be positive

    The constraints are stored in a vector (b) and matrix (A).
    The sign convention is chosen such that the constraint is satisfied
    if A.x + b < eps.

    :param assemblage: The assemblage for which the constraints are calculated.
    :type assemblage: :class:`burnman.Composite`

    :returns: The constraints vector and matrix.
    :rtype: tuple
    """
    bounds = []
    n_constraints = 0
    for i, n in enumerate(assemblage.endmembers_per_phase):
        n_constraints += 1
        if n == 1:
            bounds.append(np.array([[]]))
        else:
            bounds.append(assemblage.phases[i].solution_model.endmember_occupancies)
            n_constraints += len(bounds[-1][0])

    c_vector = np.zeros((n_constraints + 2))
    c_matrix = np.zeros(
        (n_constraints + 2, assemblage.n_endmembers + 2 + n_free_compositional_vectors)
    )  # includes P, T

    c_matrix[0, 0] = -1  # P>0
    c_matrix[1, 1] = -1  # T>0

    cidx = 2  # index of current compositional constraint
    pidx = 0  # starting index of current phase
    for i, n in enumerate(assemblage.endmembers_per_phase):
        m = len(bounds[i][0])
        # The first endmember proportion is not a free variable
        # (all endmembers in any solution must sum to one)
        # Re-express the constraints without the first endmember
        c_matrix[cidx, pidx + 2] = -1.0  # need phase proportions > 0
        cidx += 1
        if m != 0:
            c_vector[cidx : cidx + m] = -bounds[i][0]
            c_matrix[cidx : cidx + m, pidx + 1 + 2 : pidx + n + 2] = (
                np.einsum("i, j", bounds[i][0], np.ones_like(bounds[i][1:, 0]))
                - bounds[i].T[:, 1:]
            )
            cidx += m
        pidx += n

    return c_vector, c_matrix


def get_parameters(assemblage, n_free_compositional_vectors=0):
    """
    Gets the starting parameters vector (x) for the current equilibrium problem.
    These are:

      - pressure
      - temperature
      - absolute amount of each phase. if a phase is a solution
        with >1 endmember, the following parameters are the mole fractions
        of the independent endmembers in the solution, except for the first
        endmember (as the mole fractions must sum to one).

    :param assemblage: The assemblage to be equilibrated.
    :type assemblage: :class:`burnman.Composite`

    :returns: The current values of all the parameters.
    :rtype: numpy.array
    """
    params = np.zeros(assemblage.n_endmembers + 2 + n_free_compositional_vectors)
    n_moles_phase = assemblage.n_moles * np.array(assemblage.molar_fractions)

    try:
        params[:2] = [assemblage.pressure, assemblage.temperature]
    except AttributeError:
        raise Exception("You need to set_state before getting parameters")

    j = 2
    for i, ph in enumerate(assemblage.phases):
        params[j] = n_moles_phase[i]
        if isinstance(ph, Solution):
            params[j + 1 : j + assemblage.endmembers_per_phase[i]] = assemblage.phases[
                i
            ].molar_fractions[1:]
        j += assemblage.endmembers_per_phase[i]

    return params


def get_endmember_amounts(assemblage):
    """
    Gets the absolute amounts of all the endmembers in the solution.

    :param assemblage: The assemblage to be equilibrated.
    :type assemblage: :class:`burnman.Composite`

    :returns: The current amounts of all the endmembers.
    :rtype: numpy.array
    """
    phase_amounts = assemblage.n_moles * assemblage.molar_fractions
    amounts = np.empty(assemblage.n_endmembers)
    j = 0
    for i, ph in enumerate(assemblage.phases):
        if isinstance(ph, Solution):
            amounts[j : j + assemblage.endmembers_per_phase[i]] = (
                phase_amounts[i] * assemblage.phases[i].molar_fractions
            )
        else:
            amounts[j] = phase_amounts[i]
        j += assemblage.endmembers_per_phase[i]

    return amounts


def set_compositions_and_state_from_parameters(assemblage, parameters):
    """
    Sets the phase compositions, amounts and state of the assemblage
    from a list of parameter values.

    :param assemblage: The assemblage to be equilibrated.
    :type assemblage: :class:`burnman.Composite`

    :param parameters: The current parameter values.
    :type parameters: numpy.array
    """
    assemblage.set_state(parameters[0], parameters[1])
    i = 2
    phase_amounts = np.zeros(len(assemblage.phases))
    for phase_idx, ph in enumerate(assemblage.phases):
        phase_amounts[phase_idx] = parameters[i]
        if isinstance(ph, Solution):
            n_mbrs = len(ph.endmembers)
            f = [0.0] * n_mbrs
            f[1:] = parameters[i + 1 : i + n_mbrs]
            f[0] = 1.0 - sum(f)
            ph.set_composition(f)
            i += n_mbrs
        else:
            i += 1

    assert np.all(phase_amounts > -1.0e-8)
    phase_amounts = np.abs(phase_amounts)
    assemblage.n_moles = sum(phase_amounts)
    assemblage.set_fractions(phase_amounts / assemblage.n_moles)
    return None


def F(
    x,
    assemblage,
    equality_constraints,
    reduced_composition_vector,
    reduced_free_composition_vectors,
):
    """
    The vector-valued function for which the root is sought.
    The first two vector values depend on the
    equality_constraints chosen. For example, if
      - eq[i][0] = 'P', F[i] = P - eq[i][1]
      - eq[i][0] = 'T', F[i] = T - eq[i][1]
      - eq[i][0] = 'S', F[i] = entropy - eq[i][1]
      - eq[i][0] = 'V', F[i] = volume - eq[i][1]
      - eq[i][0] = 'PT_ellipse', F[i] = norm(([P, T] - eq[i][1][0])/eq[i][1][1]) - 1
      - eq[i][0] = 'X', np.dot(eq[i][1][0], x) - eq[i][1][1]

    The next set of vector values correspond to the reaction affinities.
    The final set of vector values correspond to the bulk
    composition constraints.

    :param x: Parameter values for the equilibrium problem to be solved.
    :type x: numpy array

    :param assemblage: The assemblage to be equilibrated.
    :type assemblage: :class:`burnman.Composite`

    :param equality_constraints: A list of the equality constraints
        (see above for valid formats).
    :type equality_constraints: list of lists

    :param reduced_composition_vector: The amounts of the independent
        elements.
    :type reduced_composition_vector: numpy.array

    :param reduced_free_composition_vectors: The amounts of the
        independent elements in each of the free_compositional_vectors.
    :type reduced_free_composition_vectors: 2D numpy.array

    :returns: The vector corresponding to F(x).
    :rtype: numpy.array
    """

    set_compositions_and_state_from_parameters(assemblage, x)
    new_endmember_amounts = get_endmember_amounts(assemblage)

    # We want to find the root of the following equations
    n_equality_constraints = len(equality_constraints)
    eqns = np.zeros((assemblage.n_endmembers + n_equality_constraints))
    i = 0
    for i, (type_c, eq_c) in enumerate(equality_constraints):
        if type_c == "P":
            eqns[i] = x[0] - eq_c
        elif type_c == "T":
            eqns[i] = x[1] - eq_c
        elif type_c == "S":
            eqns[i] = assemblage.molar_entropy * assemblage.n_moles - eq_c
        elif type_c == "V":
            eqns[i] = assemblage.molar_volume * assemblage.n_moles - eq_c
        elif type_c == "PT_ellipse":
            v_scaled = (x[0:2] - eq_c[0]) / eq_c[1]
            eqns[i] = np.linalg.norm(v_scaled) - 1.0
        elif type_c == "X":
            eqns[i] = np.dot(eq_c[0], x) - eq_c[1]  # i.e. Ax = b
        else:
            raise Exception("constraint type not recognised")
    i += 1
    if n_equality_constraints > 2:
        new_reduced_composition_vector = reduced_composition_vector + x[
            2 - n_equality_constraints :
        ].dot(reduced_free_composition_vectors)
    else:
        new_reduced_composition_vector = reduced_composition_vector
    eqns[i : i + assemblage.n_reactions] = assemblage.reaction_affinities
    eqns[i + assemblage.n_reactions :] = (
        np.dot(assemblage.reduced_stoichiometric_array.T, new_endmember_amounts)
        - new_reduced_composition_vector
    )
    return eqns


def jacobian(x, assemblage, equality_constraints, reduced_free_composition_vectors):
    """
    The Jacobian of the vector-valued function :math:`F` for which the
    root is sought (:math:`\\partial F / \\partial x`).
    See documentation for :func:`F` and :func:`get_parameters`
    (which return :math:`F` and :math:`x` respectively) for more details.

    :param x: Parameter values for the equilibrium problem to be solved.
    :type x: numpy.array

    :param assemblage: The assemblage to be equilibrated.
    :type assemblage: :class:`burnman.Composite`

    :param equality_constraints: A list of the equality constraints
        (see documentation for :func:`burnman.tools.equilbration.F`).
    :type equality_constraints: list of lists

    :param reduced_free_composition_vectors: The amounts of the
        independent elements in each of the free_compositional_vectors.
    :type reduced_free_composition_vectors: 2D numpy array

    :returns: The Jacobian for the equilibrium problem.
    :rtype: 2D numpy.array

    """
    # The solver always calls the Jacobian with the same
    # x parameters as used previously for the root functions
    # Therefore we don't need to set compositions or state again here.

    # First, we find out the effect of the two constraint parameters F[:2]
    # on the pressure (x[0]) and temperature (x[1]):
    n_equality_constraints = len(equality_constraints)
    jacobian = np.zeros(
        (
            assemblage.n_endmembers + n_equality_constraints,
            assemblage.n_endmembers + n_equality_constraints,
        )
    )
    ic = 0
    for ic, (type_c, eq_c) in enumerate(equality_constraints):
        if type_c == "P":  # dP/dx
            jacobian[ic, 0] = 1.0  # jacobian[i, j!=0] = 0
        elif type_c == "T":  # dT/dx
            jacobian[ic, 1] = 1.0  # jacobian[i, j!=1] = 0
        elif type_c == "S":  # dS/dx
            # dS/dP = -aV, dS/dT = Cp/T
            jacobian[ic, 0:2] = [
                -assemblage.n_moles * assemblage.alpha * assemblage.molar_volume,
                assemblage.n_moles * assemblage.molar_heat_capacity_p / x[1],
            ]
            j = 2
            for k, n in enumerate(assemblage.endmembers_per_phase):
                jacobian[ic, j] = assemblage.phases[k].molar_entropy
                if n > 1:  # for solutions with >1 endmember
                    jacobian[ic, j + 1 : j + n] = (
                        assemblage.n_moles
                        * assemblage.molar_fractions[k]
                        * (
                            assemblage.phases[k].partial_entropies[1:]
                            - assemblage.phases[k].partial_entropies[0]
                        )
                    )
                j += n
        elif type_c == "V":  # dV/dx
            # dV/dP = -V/K_T, dV/dT = aV
            jacobian[ic, 0:2] = [
                -assemblage.n_moles
                * assemblage.molar_volume
                / assemblage.isothermal_bulk_modulus_reuss,
                assemblage.n_moles * assemblage.molar_volume,
            ]
            j = 2
            for k, n in enumerate(assemblage.endmembers_per_phase):
                jacobian[ic, j] = assemblage.phases[k].molar_volume
                if n > 1:  # for solutions with >1 stable endmember
                    jacobian[ic, j + 1 : j + n] = (
                        assemblage.n_moles
                        * assemblage.molar_fractions[k]
                        * (
                            assemblage.phases[k].partial_volumes[1:]
                            - assemblage.phases[k].partial_volumes[0]
                        )
                    )
                j += n
        elif type_c == "PT_ellipse":
            v_scaled = (x[0:2] - eq_c[0]) / eq_c[1]
            jacobian[ic, 0:2] = v_scaled / (np.linalg.norm(v_scaled) * eq_c[1])
        elif type_c == "X":
            jacobian[ic, :] = eq_c[0]
        else:
            raise Exception("constraint type not recognised")
    ic += 1

    # Next, let's get the effect of pressure and temperature
    # on each of the independent reactions
    # i.e. dF(i, reactions)/dx[0] and dF(i, reactions)/dx[1]
    partial_volumes_vector = np.zeros((assemblage.n_endmembers))
    partial_entropies_vector = np.zeros((assemblage.n_endmembers))
    j = 0
    for i, n in enumerate(assemblage.endmembers_per_phase):
        if n == 1:  # for endmembers
            partial_volumes_vector[j] = assemblage.phases[i].molar_volume
            partial_entropies_vector[j] = assemblage.phases[i].molar_entropy
        else:  # for solutions
            partial_volumes_vector[j : j + n] = assemblage.phases[i].partial_volumes
            partial_entropies_vector[j : j + n] = assemblage.phases[i].partial_entropies
        j += n
    reaction_volumes = np.dot(assemblage.reaction_basis, partial_volumes_vector)
    reaction_entropies = np.dot(assemblage.reaction_basis, partial_entropies_vector)

    # dGi/dP = deltaVi; dGi/dT = -deltaSi
    jacobian[ic : ic + len(reaction_volumes), 0] = reaction_volumes
    jacobian[ic : ic + len(reaction_volumes), 1] = -reaction_entropies

    # Pressure and temperature have no effect on the bulk
    # compositional constraints
    # i.e. dF(i, bulk)/dx[0] and dF(i, bulk)/dx[1] = 0

    # Finally, let's build the compositional Hessian d2G/dfidfj = dmui/dfj
    # where fj is the fraction of endmember j in a phase
    phase_amounts = np.array(assemblage.molar_fractions) * assemblage.n_moles
    comp_hessian = np.zeros((assemblage.n_endmembers, assemblage.n_endmembers))
    dfi_dxj = np.zeros((assemblage.n_endmembers, assemblage.n_endmembers))
    dpi_dxj = np.zeros((assemblage.n_endmembers, assemblage.n_endmembers))
    j = 0
    for i, n in enumerate(assemblage.endmembers_per_phase):
        if n == 1:
            # changing the amount (p) of a pure phase
            # does not change its fraction in that phase,
            # so dfi_dxj remains unchanged
            dpi_dxj[j, j] = 1.0
        else:
            comp_hessian[j : j + n, j : j + n] = assemblage.phases[i].gibbs_hessian

            # x[0] = p(phase) and x[1:] = f[1:] - f[0]
            # Therefore
            # df[0]/dx[0] = 0
            # df[0]/dx[1:] = -1
            # (because changing the fraction of any endmember
            # depletes the fraction of the first endmember)
            # df[1:]/dx[1:] = 1 on diagonal, 0 otherwise
            # (because all other fractions are independent of each other)
            dfi_dxj[j : j + n, j : j + n] = np.eye(n)
            dfi_dxj[j, j : j + n] -= 1.0
            # Total amounts of endmembers (p) are the fractions
            # multiplied by the amounts of their representative phases
            dpi_dxj[j : j + n, j : j + n] = (
                dfi_dxj[j : j + n, j : j + n] * phase_amounts[i]
            )
            # The derivative of the amount of each endmember with respect
            # to the amount of each phase is equal to the molar fractions
            # of the endmembers.
            dpi_dxj[j : j + n, j] = assemblage.phases[i].molar_fractions
        j += n

    # dfi_dxj converts the endmember hessian to the parameter hessian.
    reaction_hessian = assemblage.reaction_basis.dot(comp_hessian).dot(dfi_dxj)
    bulk_hessian = assemblage.reduced_stoichiometric_array.T.dot(dpi_dxj)

    if reaction_hessian.shape[0] > 0:
        jacobian[ic:, 2 : 2 + len(reaction_hessian[0])] = np.concatenate(
            (reaction_hessian, bulk_hessian)
        )
    else:
        jacobian[ic:, 2 : 2 + len(bulk_hessian[0])] = bulk_hessian

    if len(reduced_free_composition_vectors) > 0:
        jacobian[
            -reduced_free_composition_vectors.shape[1] :, 2 + len(reaction_hessian[0]) :
        ] = -reduced_free_composition_vectors.T

    return jacobian


def lambda_bounds(dx, x, endmembers_per_phase):
    """
    Returns the lambda bounds for the damped affine invariant modification
    to Newton's method for nonlinear problems (Deuflhard, 1974;1975;2004).

    :param dx: The proposed newton step.
    :type dx: numpy.array

    :param x: Parameter values for the equilibrium problem to be solved.
    :type x: numpy.array

    :param endmembers_per_phase: A list of the number of endmembers in each phase.
    :type endmembers_per_phase: list of int

    :returns: Minimum and maximum allowed fractions of the full newton step (dx).
    :rtype: tuple of floats
    """

    max_steps = np.ones((len(x))) * 100000.0

    # first two constraints are P and T
    max_steps[0:2] = [20.0e9, 500.0]  # biggest reasonable P and T steps

    j = 2
    for i, n in enumerate(endmembers_per_phase):
        # if the phase fraction constraint would be broken,
        # set a step that is marginally smaller
        if x[j] + dx[j] < 0.0:
            max_steps[j] = max(x[j] * 0.999, 0.001)
        max_steps[j + 1 : j + n] = [
            max(xi * 0.99, 0.01) for xi in x[j + 1 : j + n]
        ]  # maximum compositional step
        j += n

    max_lmda = min(
        [
            1.0 if step <= max_steps[i] else max_steps[i] / step
            for i, step in enumerate(np.abs(dx))
        ]
    )

    return (1.0e-8, max_lmda)


def phase_fraction_constraints(phase, assemblage, fractions, prm):
    """
    Converts a phase fraction constraint into standard linear form
    that can be processed by the root finding problem.

    We start with a single fraction or an array of fractions
    for a particular phase (:math:`n_p / \\sum n = f`).
    These are then converted into the "X" form of constraint
    by multiplying by :math:`\\sum n` and moving all terms to
    the LHS of the equation:

    :math:`-f n_0 - f n_1 - \\ldots + (1-f) n_p - \\ldots = 0`

    This form is less readable, but easier to use as a constraint
    in a nonlinear solve.

    :param phase: The phase for which the fraction is to be constrained
    :type phase: :class:`burnman.Solution` or :class:`burnman.Mineral`

    :param assemblage: The assemblage to be equilibrated.
    :type assemblage: :class:`burnman.Composite`

    :param fractions: The phase fractions to be satified at equilibrium.
    :type fractions: numpy.array

    :param prm: A tuple with attributes n_parameters
        (the number of parameters for the current equilibrium problem)
        and phase_amount_indices (the indices of the parameters that
        correspond to phase amounts).
    :type prm: namedtuple

    :returns: The phase fraction constraints.
    :rtype: list
    """
    phase_idx = assemblage.phases.index(phase)

    constraints = []
    for fraction in fractions:
        constraints.append(["X", [np.zeros((prm.n_parameters)), 0.0]])
        constraints[-1][-1][0][prm.phase_amount_indices] = -fraction
        constraints[-1][-1][0][prm.phase_amount_indices[phase_idx]] += 1.0

    return constraints


def phase_composition_constraints(phase, assemblage, constraints, prm):
    """
    Converts a phase composition constraint into standard linear form
    that can be processed by the root finding problem.

    We start with constraints in the form (site_names, n, d, v), where
    :math:`(n x)/ (d x) = v` and :math:`n` and :math:`d` are fixed vectors of
    site coefficients. So, one could for example choose a constraint
    ([Mg_A, Fe_A], [1., 0.], [1., 1.], [0.5]) which would
    correspond to equal amounts Mg and Fe on the A site.

    This function converts the user-defined vectors of site constraints
    :math:`n` and :math:`d` into vectors of endmember proportion
    constraints :math:`n'` and :math:`d'`, such that
    :math:`(n' x)/ (d' x) = v`. This is done via linear transformation
    using the site occupancy matrix provided by :class:`burnman.Solution`.
    By multiplying by the denominator, we have the following scalar
    comparison: :math:`(n' x) = v (d' x)`

    The equilibration function does not use the proportion of
    the first endmember (as the endmember proportions must sum to one),
    and so we split :math:`x`, :math:`n'` and :math:`d'` into the first
    element and following elements:
    :math:`(n'_0 x_0 + n'_i x_i) = v (d'_0 x_0 + d'_i x_i)`
    where :math:`i` is taken over all elements apart from the first.

    With some more rearranging we can express the constraint in standard
    linear form:
    :math:`(n'_0 (1 - \\sum_j x_j) + n'_i x_i) = v (d'_0 (1 - \\sum_j x_j) + d'_i x_i)`

    :math:`(n'_0 + (n'_i - 1_i n'_0) x_i) = v (d'_0 + (d'_i - 1_i d'_0) x_i)`

    :math:`(((n'_i - 1_i n'_0) - v(d'_i - 1_i d'_0)) x_i) = (v d'_0 - n'_0)`

    This form is less readable, but easier to use as a constraint
    in a nonlinear solve.

    :param phase: The phase for which the composition is to be constrained.
    :type phase: :class:`burnman.Solution`

    :param assemblage: The assemblage to be equilibrated.
    :type assemblage: :class:`burnman.Composite`

    :param constraints: The desired constraints in the form:
        site_names (list of strings), numerator (numpy.array),
        denominator (numpy.array), values (numpy.array).
    :type constraints: tuple

    :returns: The phase composition constraints in standard form.
    :rtype: list
    """
    phase_idx = assemblage.phases.index(phase)

    site_names, numerator, denominator, values = constraints
    site_indices = [phase.solution_model.site_names.index(name) for name in site_names]
    noccs = phase.solution_model.endmember_noccupancies

    # Converts site constraints into endmember constraints
    # Ends up with shape (n_endmembers, 2)
    endmembers = np.dot(noccs[:, site_indices], np.array([numerator, denominator]).T)

    numer0, denom0 = endmembers[0]
    endmembers -= endmembers[0]
    numer, denom = endmembers.T[:, 1:]

    # We start from the correct index
    start_idx = sum(assemblage.endmembers_per_phase[:phase_idx]) + 3
    n_indices = assemblage.endmembers_per_phase[phase_idx] - 1

    x_constraints = []
    for v in values:
        f = v * denom0 - numer0
        x_constraints.append(["X", [np.zeros((prm.n_parameters)), f]])
        x_constraints[-1][1][0][start_idx : start_idx + n_indices] = numer - v * denom

    return x_constraints


def get_equilibration_parameters(assemblage, composition, free_compositional_vectors):
    """
    Builds a named tuple containing the parameter names and
    various other parameters needed by the equilibrium solve.

    :param assemblage: The assemblage to be equilibrated.
    :type assemblage: :class:`burnman.Composite`

    :param composition: The bulk composition for the equilibrium problem.
    :type composition: dict

    :param free_compositional_vectors: The bulk compositional
        degrees of freedom for the equilibrium problem.
    :type free_compositional_vectors: list of dictionaries

    :returns: A tuple with attributes n_parameters
        (the number of parameters for the current equilibrium problem)
        and phase_amount_indices (the indices of the parameters that
        correspond to phase amounts).
    :rtype: namedtuple
    """
    # Initialize a named tuple for the equilibration parameters
    prm = namedtuple("assemblage_parameters", [])

    # Process parameter names
    prm.parameter_names = ["Pressure (Pa)", "Temperature (K)"]
    for i, n in enumerate(assemblage.endmembers_per_phase):
        prm.parameter_names.append("x({0})".format(assemblage.phases[i].name))
        if n > 1:
            p_names = [
                "p({0} in {1})".format(n, assemblage.phases[i].name)
                for n in assemblage.phases[i].endmember_names[1:]
            ]
            prm.parameter_names.extend(p_names)

    n_free_compositional_vectors = len(free_compositional_vectors)
    for i in range(n_free_compositional_vectors):
        prm.parameter_names.append(f"v_{i}")

    prm.n_parameters = len(prm.parameter_names)
    prm.phase_amount_indices = [
        i for i in range(len(prm.parameter_names)) if "x(" in prm.parameter_names[i]
    ]

    # Find the bulk composition vector
    prm.bulk_composition_vector = np.array(
        [composition[e] for e in assemblage.elements]
    )

    if n_free_compositional_vectors > 0:
        prm.free_compositional_vectors = np.array(
            [
                [
                    (
                        free_compositional_vectors[i][e]
                        if e in free_compositional_vectors[i]
                        else 0.0
                    )
                    for e in assemblage.elements
                ]
                for i in range(n_free_compositional_vectors)
            ]
        )
    else:
        prm.free_compositional_vectors = np.empty((0, len(assemblage.elements)))

    if assemblage.compositional_null_basis.shape[0] != 0:
        if (
            np.abs(
                assemblage.compositional_null_basis.dot(prm.bulk_composition_vector)[0]
            )
            > 1.0e-12
        ):
            raise Exception(
                "The bulk composition is not within the "
                "compositional space of the assemblage"
            )

    prm.reduced_composition_vector = prm.bulk_composition_vector[
        assemblage.independent_element_indices
    ]
    prm.reduced_free_composition_vectors = prm.free_compositional_vectors[
        :, assemblage.independent_element_indices
    ]
    prm.constraint_vector, prm.constraint_matrix = calculate_constraints(
        assemblage, n_free_compositional_vectors
    )
    return prm


def process_eq_constraints(equality_constraints, assemblage, prm):
    """
    A function that processes the equality constraints
    into a form that can be processed by the F and jacobian functions.

    This function has two main tasks: it turns phase_fraction and
    phase_composition constraints into standard linear constraints in the
    solution parameters. It also turns vector-valued constraints into a
    list of scalar-valued constraints.

    :param equality_constraints: A list of equality constraints.
        For valid types of constraints, please see the documentation for
        :func:`burnman.equilibrate`.
    :type equality_constraints: list

    :param assemblage: The assemblage to be equilibrated.
    :type assemblage: :class:`burnman.Composite`

    :param prm: A tuple with attributes n_parameters
        (the number of parameters for the current equilibrium problem)
        and phase_amount_indices (the indices of the parameters that
        correspond to phase amounts).
    :type prm: namedtuple

    :returns: Equality constraints in a form that can be processed
        by the F and jacobian functions.
    :rtype: list of lists
    """
    eq_constraint_lists = []
    for i in range(len(equality_constraints)):
        if equality_constraints[i][0] == "phase_fraction":
            phase = equality_constraints[i][1][0]
            fraction = equality_constraints[i][1][1]
            if isinstance(fraction, float):
                fraction = np.array([fraction])
            if not isinstance(fraction, np.ndarray):
                raise Exception(
                    "The constraint fraction in equality {0} "
                    "should be a float or numpy array".format(i + 1)
                )
            eq_constraint_lists.append(
                phase_fraction_constraints(phase, assemblage, fraction, prm)
            )
        elif equality_constraints[i][0] == "phase_composition":
            phase = equality_constraints[i][1][0]
            constraint = equality_constraints[i][1][1]
            if isinstance(constraint[3], float):
                constraint = (
                    constraint[0],
                    constraint[1],
                    constraint[2],
                    np.array([constraint[3]]),
                )
            if not isinstance(constraint[3], np.ndarray):
                raise Exception(
                    "The last constraint parameter in equality "
                    "{0} should be a float "
                    "or numpy array".format(i + 1)
                )

            eq_constraint_lists.append(
                phase_composition_constraints(phase, assemblage, constraint, prm)
            )
        elif equality_constraints[i][0] == "X":
            constraint = equality_constraints[i][1]
            if isinstance(constraint[-1], float):
                constraint = (constraint[0], np.array([constraint[-1]]))
            if not isinstance(constraint[-1], np.ndarray):
                raise Exception(
                    "The last constraint parameter in "
                    "equality {0} should be "
                    "a float or numpy array".format(i + 1)
                )

            eq_constraint_lists.append(
                [["X", [constraint[0], p]] for p in constraint[1]]
            )

        elif (
            equality_constraints[i][0] == "P"
            or equality_constraints[i][0] == "T"
            or equality_constraints[i][0] == "PT_ellipse"
            or equality_constraints[i][0] == "S"
            or equality_constraints[i][0] == "V"
        ):
            if isinstance(equality_constraints[i][1], float):
                equality_constraints[i] = (
                    equality_constraints[i][0],
                    np.array([equality_constraints[i][1]]),
                )
            if not isinstance(equality_constraints[i][1], np.ndarray):
                raise Exception(
                    "The last parameter in "
                    f"equality_constraint[{i+1}] should be a "
                    "float or numpy array"
                )
            eq_constraint_lists.append(
                [[equality_constraints[i][0], p] for p in equality_constraints[i][1]]
            )
        else:
            raise Exception(
                "The type of equality_constraint is "
                "not recognised for constraint {0}.\n"
                "Should be one of P, T, S, V, X,\n"
                "PT_ellipse, phase_fraction, "
                "or phase_composition.".format(i + 1)
            )
    return eq_constraint_lists


def equilibrate(
    composition,
    assemblage,
    equality_constraints,
    free_compositional_vectors=[],
    tol=1.0e-3,
    store_iterates=False,
    store_assemblage=True,
    max_iterations=100.0,
    verbose=False,
):
    """
    A function that finds the thermodynamic equilibrium state of an
    assemblage subject to given equality constraints by solving a set of
    nonlinear equations related to the chemical potentials and
    other state variables of the system.

    The user chooses an assemblage (e.g. olivine, garnet and orthopyroxene)
    and :math:`2+n_c` equality constraints, where :math:`n_c` is the number of
    bulk compositional degrees of freedom. The equilibrate function attempts to
    find the remaining unknowns that satisfy those constraints.

    There are a number of equality constraints implemented in burnman. These are
    given as a list of lists. Each constraint should have the form:
    [<constraint type>, <constraint>], where
    <constraint type> is one of 'P', 'T', 'S', 'V', 'X', 'PT_ellipse',
    'phase_fraction', or 'phase_composition'. The format of the
    <constraint> object depends on the constraint type:
        - P: float or numpy.array of
            pressures [Pa]
        - T: float or numpy.array of
            temperatures [K]
        - S: float or numpy.array of
            entropies [J/K]
        - V: float or numpy.array of
            volumes [m:math:`^3`]
        - PT_ellipse: list of two floats or numpy.arrays, where the equality
            satifies the equation norm(([P, T] - arr[0])/arr[1]) = 1
        - phase_fraction: tuple with the form (<phase> <fraction(s)>),
            where <phase> is one of the phase objects in the assemblage
            and <fraction(s)> is a float or numpy.array corresponding
            to the desired phase fractions.
        - phase_composition: tuple with the form (<phase> <constraint>),
            where <phase> is one of the phase objects in the assemblage
            and <constraint> has the form (site_names, n, d, v),
            where :math:`(nx)/(dx) = v`, n and d are constant vectors of
            site coefficients, and v is a float or numpy.array. For example,
            a constraint of the form ([Mg_A, Fe_A], [1., 0.], [1., 1.], [0.5])
            would correspond to equal amounts Mg and Fe on the A site
            (Mg_A / (Mg_A + Fe_A) = 0.5).
        - X: list of a numpy.array and a float or numpy.array,
            where the equality satisfies the linear equation arr[0] x = arr[1].
            The first numpy.array is fixed, and the second is to be looped over
            by the equilibrate function.
            This is a generic compositional equality constraint.

    :param composition: The bulk composition that the assemblage must satisfy.
    :type composition: dict

    :param assemblage: The assemblage to be equilibrated.
    :type assemblage: :class:`burnman.Composite`

    :param equality_constraints: The list of equality constraints. See above
        for valid formats.
    :type equality_constraints: list of list

    :param free_compositional_vectors: A list of dictionaries containing
        the compositional freedom of the solution. For example, if the list
        contains the vector {'Mg': 1., 'Fe': -1}, that implies that the bulk
        composition is equal to composition + :math:`a` (n_Mg - n_Fe),
        where the value of :math:`a` is to be determined by the solve.
        Vector given in atomic (molar) units of elements.
    :type free_compositional_vectors: list of dict

    :param tol: The tolerance for the nonlinear solver.
    :type tol: float

    :param store_iterates: Whether to store the parameter values for
        each iteration in each solution object.
    :type store_iterates: bool

    :param store_assemblage: Whether to store a copy of the assemblage
        object in each solution object.
    :type store_assemblage: bool

    :param max_iterations: The maximum number of iterations for the
        nonlinear solver.
    :type max_iterations: int

    :param verbose: Whether to print output updating the user on the status of
        equilibration.
    :type verbose: bool

    :returns: Solver solution object (or a list, or a 2D list of solution objects)
        created by :func:`burnman.optimize.nonlinear_solvers.damped_newton_solve`,
        and a namedtuple object created by
        :func:`burnman.tools.equilibration.get_equilibration_parameters`. See
        documentation of these functions for the return types. If store_assemblage
        is True, each solution object also has an attribute called `assemblage`,
        which contains a copy of the input assemblage with the equilibrated
        properties. So, for a 2D grid of solution objects, one could call either
        sols[0][1].x[0] or sols[0][1].assemblage.pressure to get the pressure.
    :rtype: tuple
    """
    for ph in assemblage.phases:
        if isinstance(ph, Solution) and not hasattr(ph, "molar_fractions"):
            raise Exception(
                f"set_composition for solution {ph} before running equilibrate."
            )

    if assemblage.molar_fractions is None:
        n_phases = len(assemblage.phases)
        f = 1.0 / float(n_phases)
        assemblage.set_fractions([f for i in range(n_phases)])

    assemblage.n_moles = sum(composition.values()) / sum(assemblage.formula.values())

    n_equality_constraints = len(equality_constraints)
    n_free_compositional_vectors = len(free_compositional_vectors)

    if n_equality_constraints != n_free_compositional_vectors + 2:
        raise Exception(
            "The number of equality constraints "
            f"(currently {n_equality_constraints}) "
            "must be two more than the number of "
            "free_compositional vectors "
            f"(currently {n_free_compositional_vectors})."
        )

    for v in free_compositional_vectors:
        if np.abs(sum(v.values())) > 1.0e-12:
            raise Exception(
                "The amounts of each free_compositional_vector" "must sum to zero"
            )

    # Make parameter tuple
    prm = get_equilibration_parameters(
        assemblage, composition, free_compositional_vectors
    )

    # Check equality constraints have the correct structure
    # Convert into the format readable by the function and jacobian functions
    eq_constraint_lists = process_eq_constraints(equality_constraints, assemblage, prm)

    # Set up solves
    nc = [len(eq_constraint_list) for eq_constraint_list in eq_constraint_lists]

    # Find the initial state (could be none here)
    initial_state = [assemblage.pressure, assemblage.temperature]

    # Reset initial state if equality constraints
    # are related to pressure or temperature
    for i in range(n_equality_constraints):
        if eq_constraint_lists[i][0][0] == "P":
            initial_state[0] = eq_constraint_lists[i][0][1]
        elif eq_constraint_lists[i][0][0] == "T":
            initial_state[1] = eq_constraint_lists[i][0][1]
        elif eq_constraint_lists[i][0][0] == "PT_ellipse":
            initial_state = eq_constraint_lists[i][0][1][1]

    if initial_state[0] is None:
        initial_state[0] = 5.0e9
    if initial_state[1] is None:
        initial_state[1] = 1200.0

    assemblage.set_state(*initial_state)
    parameters = get_parameters(assemblage, n_free_compositional_vectors)

    # Solve the system of equations, loop over input parameters
    sol_array = np.empty(shape=tuple(nc), dtype="object")

    # Loop over problems
    problems = list(product(*[list(range(nc[i])) for i in range(len(nc))]))
    n_problems = len(problems)
    for i_problem, i_c in enumerate(problems):
        if verbose:
            string = "Processing solution"
            for i in range(len(i_c)):
                string += " {0}/{1}".format(i_c[i] + 1, nc[i])

            print(string + ":")

        equality_constraints = [eq_constraint_lists[i][i_c[i]] for i in range(len(nc))]

        # Set the initial fractions and compositions
        # of the phases in the assemblage:
        sol = damped_newton_solve(
            F=lambda x: F(
                x,
                assemblage,
                equality_constraints,
                prm.reduced_composition_vector,
                prm.reduced_free_composition_vectors,
            ),
            J=lambda x: jacobian(
                x,
                assemblage,
                equality_constraints,
                prm.reduced_free_composition_vectors,
            ),
            lambda_bounds=lambda dx, x: lambda_bounds(
                dx, x, assemblage.endmembers_per_phase
            ),
            guess=parameters,
            linear_constraints=(prm.constraint_matrix, prm.constraint_vector),
            tol=tol,
            store_iterates=store_iterates,
            max_iterations=max_iterations,
        )

        if sol.success and len(assemblage.reaction_affinities) > 0.0:
            maxres = np.max(np.abs(assemblage.reaction_affinities)) + 1.0e-5
            assemblage.equilibrium_tolerance = maxres

        if store_assemblage:
            sol.assemblage = assemblage.copy()
            if sol.success and len(assemblage.reaction_affinities) > 0.0:
                sol.assemblage.equilibrium_tolerance = maxres

        if verbose:
            print(sol.text)

        sol_array[i_c] = sol

        # Next, we use the solution values and Jacobian
        # to provide a starting guess for the next problem.
        # First, we find the equality constraints for the next problem
        if i_problem < n_problems - 1:
            next_i_c = problems[i_problem + 1]

            next_equality_constraints = [
                eq_constraint_lists[i][next_i_c[i]] for i in range(len(nc))
            ]

            # We use the nearest solutions as potential starting points
            # to make the next guess
            prev_sols = []
            for i in range(len(nc)):
                if next_i_c[i] != 0:
                    prev_i_c = np.copy(next_i_c)
                    prev_i_c[i] -= 1
                    prev_sols.append(sol_array[tuple(prev_i_c)])

            updated_params = False
            for s in prev_sols:
                if s.success and not updated_params:
                    # next guess based on a Newton step
                    # using the old solution vector and Jacobian
                    # with the new constraints.
                    dF = F(
                        s.x,
                        assemblage,
                        next_equality_constraints,
                        prm.reduced_composition_vector,
                        prm.reduced_free_composition_vectors,
                    )
                    luJ = lu_factor(s.J)
                    new_parameters = s.x + lu_solve(luJ, -dF)
                    c = (
                        prm.constraint_matrix.dot(new_parameters)
                        + prm.constraint_vector
                    )
                    if all(c <= 0.0):  # accept new guess
                        parameters = new_parameters
                    else:  # use the parameters from this step
                        parameters = s.x
                        exhausted_phases = [
                            assemblage.phases[phase_idx].name
                            for phase_idx, v in enumerate(
                                new_parameters[prm.phase_amount_indices]
                            )
                            if v < 0.0
                        ]
                        if len(exhausted_phases) > 0 and verbose:
                            print(
                                "A phase might be exhausted before the "
                                f"next step: {exhausted_phases}"
                            )

                    updated_params = True

    # Finally, make dimensions of sol_array equal the input dimensions
    if np.prod(sol_array.shape) > 1:
        sol_array = np.squeeze(sol_array)
    else:
        sol_array = sol_array.flatten()[0]

    return sol_array, prm



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.


"""
Matas_etal_2007
^^^^^^^^^^^^^^^

Minerals from Matas et al. 2007 and references therein. See Table 1 and 2.
"""
from __future__ import absolute_import

from ..classes.mineral import Mineral


class mg_perovskite(Mineral):
    def __init__(self):
        self.params = {
            "equation_of_state": "mgd2",
            "V_0": 24.43e-6,
            "K_0": 250.0e9,
            "Kprime_0": 4.0,
            "G_0": 175.0e9,
            "Gprime_0": 1.8,
            "molar_mass": 0.1020,
            "n": 5,
            "Debye_0": 1070.0,
            "grueneisen_0": 1.48,
            "q_0": 1.4,
        }

        Mineral.__init__(self)


class fe_perovskite(Mineral):
    def __init__(self):
        self.params = {
            "equation_of_state": "mgd2",
            "V_0": 25.34e-6,
            "K_0": 250.0e9,
            "Kprime_0": 4.0,
            "G_0": 135.0e9,
            "Gprime_0": 1.3,
            "molar_mass": 0.1319,
            "n": 5,
            "Debye_0": 841.0,
            "grueneisen_0": 1.48,
            "q_0": 1.4,
        }

        Mineral.__init__(self)


class al_perovskite(Mineral):
    def __init__(self):
        self.params = {
            "equation_of_state": "mgd2",
            "V_0": 24.58e-6,
            "K_0": 249.0e9,
            "Kprime_0": 4.0,
            "G_0": 165.0e9,
            "Gprime_0": 1.8,
            "molar_mass": 0.1005,
            "n": 5,
            "Debye_0": 1021.0,
            "grueneisen_0": 1.48,
            "q_0": 1.4,
        }

        Mineral.__init__(self)


class ca_perovskite(Mineral):
    def __init__(self):
        self.params = {
            "equation_of_state": "mgd2",
            "V_0": 27.45e-6,
            "K_0": 236.0e9,
            "Kprime_0": 3.9,
            "G_0": 165.0e9,
            "Gprime_0": 2.46,
            "molar_mass": 0.11616,
            "n": 5,
            "Debye_0": 984.0,
            "grueneisen_0": 1.53,
            "q_0": 1.6,
        }

        Mineral.__init__(self)


class periclase(Mineral):
    def __init__(self):
        self.params = {
            "equation_of_state": "mgd2",
            "V_0": 11.25e-6,
            "K_0": 160.1e9,
            "Kprime_0": 3.83,
            "G_0": 130.0e9,
            "Gprime_0": 2.2,
            "molar_mass": 0.0403,
            "n": 2,
            "Debye_0": 673.0,
            "grueneisen_0": 1.41,
            "q_0": 1.3,
        }

        Mineral.__init__(self)


class wuestite(Mineral):
    def __init__(self):
        self.params = {
            "equation_of_state": "mgd2",
            "V_0": 12.26e-6,
            "K_0": 160.1e9,
            "Kprime_0": 3.83,
            "G_0": 46.0e9,
            "Gprime_0": 0.6,
            "molar_mass": 0.0718,
            "n": 2,
            "Debye_0": 673.0,
            "grueneisen_0": 1.41,
            "q_0": 1.3,
        }

        Mineral.__init__(self)


ca_bridgmanite = ca_perovskite
mg_bridgmanite = mg_perovskite
fe_bridgmanite = fe_perovskite
al_bridgmanite = al_perovskite



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2024 by the BurnMan team, released under the GNU
# GPL v2 or later.

import numpy as np
from burnman.eos.birch_murnaghan import BirchMurnaghanBase as BM3
from burnman.classes.calibrant import Calibrant
from burnman.utils.unitcell import molar_volume_from_unit_cell_volume

"""
Anderson_1989
^^^^^^^^^^^^^
"""


class Au(Calibrant):
    """
    The Au pressure standard reported by
    Anderson et al. (1989; https://doi.org/10.1063/1.342969).
    """

    def __init__(self):
        def _pressure(volume, temperature, params):

            # Isothermal pressure (GPa)
            pressure_model = BM3()
            P0 = pressure_model.pressure(params["T_0"], volume, params)

            # Thermal pressure
            Pth_rel = (7.14e6 + params["dKTdT"] * np.log(params["V_0"] / volume)) * (
                temperature - 300.0
            )

            return P0 + Pth_rel

        Z = 4.0
        _params = {
            "V_0": molar_volume_from_unit_cell_volume(67.850, Z),
            "K_0": 166.65e9,
            "Kprime_0": 5.4823,
            "dKTdT": -11.5e6,
            "n": 1.0,
            "T_0": 300.0,
            "P_0": 0.0,
            "Z": Z,
        }

        Calibrant.__init__(self, _pressure, "pressure", _params)
